@@ -1,7 +1,9 @@
 /* smbk5pwd.c - Overlay for managing Samba and Heimdal passwords */
-/* $OpenLDAP: pkg/ldap/contrib/slapd-modules/smbk5pwd/smbk5pwd.c,v 1.17.2.3 2007/08/31 23:13:52 quanah Exp $ */
-/*
- * Copyright 2004-2005 by Howard Chu, Symas Corp.
+/* $OpenLDAP$ */
+/* This work is part of OpenLDAP Software <http://www.openldap.org/>.
+ *
+ * Copyright 2004-2021 The OpenLDAP Foundation.
+ * Portions Copyright 2004-2005 by Howard Chu, Symas Corp.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -12,11 +14,10 @@
  * top-level directory of the distribution or, alternatively, at
  * <http://www.OpenLDAP.org/license.html>.
  */
-/*
+/* ACKNOWLEDGEMENTS:
  * Support for table-driven configuration added by Pierangelo Masarati.
  * Support for sambaPwdMustChange and sambaPwdCanChange added by Marco D'Ettorre.
- *
- * The conditions of the OpenLDAP Public License apply.
+ * Support for shadowLastChange added by SATOH Fumiyasu @ OSS Technology, Inc.
  */
 
 #include <portable.h>
@@ -31,7 +32,7 @@
 #include <ac/errno.h>
 #include <ac/string.h>
 
-#include "config.h"
+#include "slap-config.h"
 
 #ifdef DO_KRB5
 #include <lber.h>
@@ -59,15 +60,20 @@ static HDB *db;
 static AttributeDescription *ad_krb5Key;
 static AttributeDescription *ad_krb5KeyVersionNumber;
 static AttributeDescription *ad_krb5PrincipalName;
+static AttributeDescription *ad_krb5ValidEnd;
 static ObjectClass *oc_krb5KDCEntry;
 #endif
 
 #ifdef DO_SAMBA
-#include <openssl/des.h>
+#ifdef HAVE_GNUTLS
+#include <nettle/md4.h>
+#elif HAVE_OPENSSL
 #include <openssl/md4.h>
+#else
+#error Unsupported crypto backend.
+#endif
 #include "ldap_utf8.h"
 
-static AttributeDescription *ad_sambaLMPassword;
 static AttributeDescription *ad_sambaNTPassword;
 static AttributeDescription *ad_sambaPwdLastSet;
 static AttributeDescription *ad_sambaPwdMustChange;
@@ -75,14 +81,21 @@ static AttributeDescription *ad_sambaPwdCanChange;
 static ObjectClass *oc_sambaSamAccount;
 #endif
 
+#ifdef DO_SHADOW
+static AttributeDescription *ad_shadowLastChange;
+static ObjectClass *oc_shadowAccount;
+#endif
+
 /* Per-instance configuration information */
 typedef struct smbk5pwd_t {
 	unsigned	mode;
 #define	SMBK5PWD_F_KRB5		(0x1U)
 #define	SMBK5PWD_F_SAMBA	(0x2U)
+#define	SMBK5PWD_F_SHADOW	(0x4U)
 
 #define SMBK5PWD_DO_KRB5(pi)	((pi)->mode & SMBK5PWD_F_KRB5)
 #define SMBK5PWD_DO_SAMBA(pi)	((pi)->mode & SMBK5PWD_F_SAMBA)
+#define SMBK5PWD_DO_SHADOW(pi)	((pi)->mode & SMBK5PWD_F_SHADOW)
 
 #ifdef DO_KRB5
 	/* nothing yet */
@@ -91,8 +104,12 @@ typedef struct smbk5pwd_t {
 #ifdef DO_SAMBA
 	/* How many seconds before forcing a password change? */
 	time_t	smb_must_change;
-        /* How many seconds after allowing a password change? */
-        time_t  smb_can_change;
+	/* How many seconds after allowing a password change? */
+	time_t  smb_can_change;
+#endif
+
+#ifdef DO_SHADOW
+	/* nothing yet */
 #endif
 } smbk5pwd_t;
 
@@ -104,6 +121,9 @@ static const unsigned SMBK5PWD_F_ALL	=
 #ifdef DO_SAMBA
 	| SMBK5PWD_F_SAMBA
 #endif
+#ifdef DO_SHADOW
+	| SMBK5PWD_F_SHADOW
+#endif
 ;
 
 static int smbk5pwd_modules_init( smbk5pwd_t *pi );
@@ -111,27 +131,6 @@ static int smbk5pwd_modules_init( smbk5pwd_t *pi );
 #ifdef DO_SAMBA
 static const char hex[] = "0123456789abcdef";
 
-/* From liblutil/passwd.c... */
-static void lmPasswd_to_key(
-	const char *lmPasswd,
-	des_cblock *key)
-{
-	const unsigned char *lpw = (const unsigned char *)lmPasswd;
-	unsigned char *k = (unsigned char *)key;
-
-	/* make room for parity bits */
-	k[0] = lpw[0];
-	k[1] = ((lpw[0]&0x01)<<7) | (lpw[1]>>1);
-	k[2] = ((lpw[1]&0x03)<<6) | (lpw[2]>>2);
-	k[3] = ((lpw[2]&0x07)<<5) | (lpw[3]>>3);
-	k[4] = ((lpw[3]&0x0F)<<4) | (lpw[4]>>4);
-	k[5] = ((lpw[4]&0x1F)<<3) | (lpw[5]>>5);
-	k[6] = ((lpw[5]&0x3F)<<2) | (lpw[6]>>6);
-	k[7] = ((lpw[6]&0x7F)<<1);
-
-	des_set_odd_parity( key );
-}
-
 #define MAX_PWLEN 256
 #define	HASHLEN	16
 
@@ -156,32 +155,6 @@ static void hexify(
 	*a++ = '\0';
 }
 
-static void lmhash(
-	struct berval *passwd,
-	struct berval *hash
-)
-{
-	char UcasePassword[15];
-	des_cblock key;
-	des_key_schedule schedule;
-	des_cblock StdText = "KGS!@#$%";
-	des_cblock hbuf[2];
-
-	strncpy( UcasePassword, passwd->bv_val, 14 );
-	UcasePassword[14] = '\0';
-	ldap_pvt_str2upper( UcasePassword );
-
-	lmPasswd_to_key( UcasePassword, &key );
-	des_set_key_unchecked( &key, schedule );
-	des_ecb_encrypt( &StdText, &hbuf[0], schedule , DES_ENCRYPT );
-
-	lmPasswd_to_key( &UcasePassword[7], &key );
-	des_set_key_unchecked( &key, schedule );
-	des_ecb_encrypt( &StdText, &hbuf[1], schedule , DES_ENCRYPT );
-
-	hexify( (char *)hbuf, hash );
-}
-
 static void nthash(
 	struct berval *passwd,
 	struct berval *hash
@@ -192,14 +165,24 @@ static void nthash(
 	 * 256 UCS2 characters, not 256 bytes...
 	 */
 	char hbuf[HASHLEN];
+#ifdef HAVE_OPENSSL
 	MD4_CTX ctx;
+#elif defined(HAVE_GNUTLS)
+	struct md4_ctx ctx;
+#endif
 
 	if (passwd->bv_len > MAX_PWLEN*2)
 		passwd->bv_len = MAX_PWLEN*2;
-		
+
+#ifdef HAVE_OPENSSL
 	MD4_Init( &ctx );
 	MD4_Update( &ctx, passwd->bv_val, passwd->bv_len );
 	MD4_Final( (unsigned char *)hbuf, &ctx );
+#elif defined(HAVE_GNUTLS)
+	md4_init( &ctx );
+	md4_update( &ctx, passwd->bv_len, (unsigned char *)passwd->bv_val );
+	md4_digest( &ctx, sizeof(hbuf), (unsigned char *)hbuf );
+#endif
 
 	hexify( hbuf, hash );
 }
@@ -215,7 +198,7 @@ static int smbk5pwd_op_cleanup(
 
 	/* clear out the current key */
 	ldap_pvt_thread_pool_setkey( op->o_threadctx, smbk5pwd_op_cleanup,
-		NULL, NULL );
+		NULL, 0, NULL, NULL );
 
 	/* free the callback */
 	cb = op->o_callback;
@@ -234,8 +217,8 @@ static int smbk5pwd_op_bind(
 	 */
 	if ( op->oq_bind.rb_method == LDAP_AUTH_SIMPLE ) {
 		slap_callback *cb;
-		ldap_pvt_thread_pool_setkey( op->o_threadctx, smbk5pwd_op_cleanup, op,
-			NULL );
+		ldap_pvt_thread_pool_setkey( op->o_threadctx,
+			smbk5pwd_op_cleanup, op, 0, NULL, NULL );
 		cb = op->o_tmpcalloc( 1, sizeof(slap_callback), op->o_tmpmemctx );
 		cb->sc_cleanup = smbk5pwd_op_cleanup;
 		cb->sc_next = op->o_callback;
@@ -268,22 +251,23 @@ static int k5key_chk(
 	const struct berval *cred,
 	const char **text )
 {
-	void *ctx;
+	void *ctx, *op_tmp;
 	Operation *op;
 	int rc;
 	Entry *e;
 	Attribute *a;
-    krb5_error_code ret;
-    krb5_keyblock key;
-    krb5_salt salt;
+	krb5_error_code ret;
+	krb5_keyblock key;
+	krb5_salt salt;
 	hdb_entry ent;
 
 	/* Find our thread context, find our Operation */
 	ctx = ldap_pvt_thread_pool_context();
 
-	if ( ldap_pvt_thread_pool_getkey( ctx, smbk5pwd_op_cleanup, (void **)&op, NULL ) ||
-		!op )
+	if ( ldap_pvt_thread_pool_getkey( ctx, smbk5pwd_op_cleanup, &op_tmp, NULL )
+		 || !op_tmp )
 		return LUTIL_PASSWD_ERR;
+	op = op_tmp;
 
 	rc = be_entry_get_rw( op, &op->o_req_ndn, NULL, NULL, 0, &e );
 	if ( rc != LDAP_SUCCESS ) return LUTIL_PASSWD_ERR;
@@ -299,6 +283,19 @@ static int k5key_chk(
 		memset( &ent, 0, sizeof(ent) );
 		ret = krb5_parse_name(context, a->a_vals[0].bv_val, &ent.principal);
 		if ( ret ) break;
+
+		a = attr_find( e->e_attrs, ad_krb5ValidEnd );
+		if (a) {
+			struct lutil_tm tm;
+			struct lutil_timet tt;
+			if ( lutil_parsetime( a->a_vals[0].bv_val, &tm ) == 0 &&
+				lutil_tm2time( &tm, &tt ) == 0 && tt.tt_usec < op->o_time ) {
+				/* Account is expired */
+				rc = LUTIL_PASSWD_ERR;
+				break;
+			}
+		}
+
 		krb5_get_pw_salt( context, ent.principal, &salt );
 		krb5_free_principal( context, ent.principal );
 
@@ -349,6 +346,7 @@ static int smbk5pwd_exop_passwd(
 	Modifications *ml;
 	slap_overinst *on = (slap_overinst *)op->o_bd->bd_info;
 	smbk5pwd_t *pi = on->on_bi.bi_private;
+	char term;
 
 	/* Not the operation we expected, pass it on... */
 	if ( ber_bvcmp( &slap_EXOP_MODIFY_PASSWD, &op->ore_reqoid ) ) {
@@ -359,12 +357,16 @@ static int smbk5pwd_exop_passwd(
 	rc = be_entry_get_rw( op, &op->o_req_ndn, NULL, NULL, 0, &e );
 	if ( rc != LDAP_SUCCESS ) return rc;
 
+	term = qpw->rs_new.bv_val[qpw->rs_new.bv_len];
+	qpw->rs_new.bv_val[qpw->rs_new.bv_len] = '\0';
+
 #ifdef DO_KRB5
 	/* Kerberos stuff */
 	do {
 		krb5_error_code ret;
 		hdb_entry ent;
 		struct berval *keys;
+		size_t nkeys;
 		int kvno, i;
 		Attribute *a;
 
@@ -392,10 +394,12 @@ static int smbk5pwd_exop_passwd(
 			/* shouldn't happen, this is a required attr */
 			Debug( LDAP_DEBUG_ANY, "%s smbk5pwd EXOP: "
 				"dn=\"%s\" missing krb5KeyVersionNumber\n",
-				op->o_log_prefix, e->e_name.bv_val, 0 );
+				op->o_log_prefix, e->e_name.bv_val );
 		}
 
-		ret = _kadm5_set_keys(kadm_context, &ent, qpw->rs_new.bv_val);
+		ret = hdb_generate_key_set_password(context, ent.principal,
+			qpw->rs_new.bv_val, &ent.keys.val, &nkeys);
+		ent.keys.len = nkeys;
 		hdb_seal_keys(context, db, &ent);
 		krb5_free_principal( context, ent.principal );
 
@@ -414,7 +418,7 @@ static int smbk5pwd_exop_passwd(
 		}
 		BER_BVZERO( &keys[i] );
 
-		_kadm5_free_keys(kadm_context, ent.keys.len, ent.keys.val);
+		hdb_free_keys(context, ent.keys.len, ent.keys.val);
 
 		if ( i != ent.keys.len ) {
 			ber_bvarray_free( keys );
@@ -431,6 +435,7 @@ static int smbk5pwd_exop_passwd(
 #ifdef SLAP_MOD_INTERNAL
 		ml->sml_flags = SLAP_MOD_INTERNAL;
 #endif
+		ml->sml_numvals = i;
 		ml->sml_values = keys;
 		ml->sml_nvalues = NULL;
 		
@@ -443,6 +448,7 @@ static int smbk5pwd_exop_passwd(
 #ifdef SLAP_MOD_INTERNAL
 		ml->sml_flags = SLAP_MOD_INTERNAL;
 #endif
+		ml->sml_numvals = 1;
 		ml->sml_values = ch_malloc( 2 * sizeof(struct berval));
 		ml->sml_values[0].bv_val = ch_malloc( 64 );
 		ml->sml_values[0].bv_len = sprintf(ml->sml_values[0].bv_val,
@@ -458,7 +464,7 @@ static int smbk5pwd_exop_passwd(
 		struct berval *keys;
 		ber_len_t j,l;
 		wchar_t *wcs, wc;
-		char *c, *d;
+		char *c;
 		struct berval pwd;
 		
 		/* Expand incoming UTF8 string to UCS4 */
@@ -492,32 +498,7 @@ static int smbk5pwd_exop_passwd(
 #ifdef SLAP_MOD_INTERNAL
 		ml->sml_flags = SLAP_MOD_INTERNAL;
 #endif
-		ml->sml_values = keys;
-		ml->sml_nvalues = NULL;
-
-		/* Truncate UCS2 to 8-bit ASCII */
-		c = pwd.bv_val+1;
-		d = pwd.bv_val+2;
-		for (j=1; j<l; j++) {
-			*c++ = *d++;
-			d++;
-		}
-		pwd.bv_len /= 2;
-		pwd.bv_val[pwd.bv_len] = '\0';
-
-		ml = ch_malloc(sizeof(Modifications));
-		ml->sml_next = qpw->rs_mods;
-		qpw->rs_mods = ml;
-
-		keys = ch_malloc( 2 * sizeof(struct berval) );
-		BER_BVZERO( &keys[1] );
-		lmhash( &pwd, keys );
-		
-		ml->sml_desc = ad_sambaLMPassword;
-		ml->sml_op = LDAP_MOD_REPLACE;
-#ifdef SLAP_MOD_INTERNAL
-		ml->sml_flags = SLAP_MOD_INTERNAL;
-#endif
+		ml->sml_numvals = 1;
 		ml->sml_values = keys;
 		ml->sml_nvalues = NULL;
 
@@ -528,9 +509,9 @@ static int smbk5pwd_exop_passwd(
 		qpw->rs_mods = ml;
 
 		keys = ch_malloc( 2 * sizeof(struct berval) );
-		keys[0].bv_val = ch_malloc( STRLENOF( "9223372036854775807L" ) + 1 );
+		keys[0].bv_val = ch_malloc( LDAP_PVT_INTTYPE_CHARS(long) );
 		keys[0].bv_len = snprintf(keys[0].bv_val,
-			STRLENOF( "9223372036854775807L" ) + 1,
+			LDAP_PVT_INTTYPE_CHARS(long),
 			"%ld", slap_get_time());
 		BER_BVZERO( &keys[1] );
 		
@@ -539,6 +520,7 @@ static int smbk5pwd_exop_passwd(
 #ifdef SLAP_MOD_INTERNAL
 		ml->sml_flags = SLAP_MOD_INTERNAL;
 #endif
+		ml->sml_numvals = 1;
 		ml->sml_values = keys;
 		ml->sml_nvalues = NULL;
 
@@ -549,9 +531,9 @@ static int smbk5pwd_exop_passwd(
 			qpw->rs_mods = ml;
 
 			keys = ch_malloc( 2 * sizeof(struct berval) );
-			keys[0].bv_val = ch_malloc( STRLENOF( "9223372036854775807L" ) + 1 );
+			keys[0].bv_val = ch_malloc( LDAP_PVT_INTTYPE_CHARS(long) );
 			keys[0].bv_len = snprintf(keys[0].bv_val,
-					STRLENOF( "9223372036854775807L" ) + 1,
+					LDAP_PVT_INTTYPE_CHARS(long),
 					"%ld", slap_get_time() + pi->smb_must_change);
 			BER_BVZERO( &keys[1] );
 
@@ -560,34 +542,66 @@ static int smbk5pwd_exop_passwd(
 #ifdef SLAP_MOD_INTERNAL
 			ml->sml_flags = SLAP_MOD_INTERNAL;
 #endif
+			ml->sml_numvals = 1;
 			ml->sml_values = keys;
 			ml->sml_nvalues = NULL;
 		}
 
-                if (pi->smb_can_change)
-                {
-                        ml = ch_malloc(sizeof(Modifications));
-                        ml->sml_next = qpw->rs_mods;
-                        qpw->rs_mods = ml;
+		if (pi->smb_can_change)
+		{
+			ml = ch_malloc(sizeof(Modifications));
+			ml->sml_next = qpw->rs_mods;
+			qpw->rs_mods = ml;
 
-                        keys = ch_malloc( 2 * sizeof(struct berval) );
-                        keys[0].bv_val = ch_malloc( STRLENOF( "9223372036854775807L" ) + 1 );
-                        keys[0].bv_len = snprintf(keys[0].bv_val,
-                                        STRLENOF( "9223372036854775807L" ) + 1,
-                                        "%ld", slap_get_time() + pi->smb_can_change);
-                        BER_BVZERO( &keys[1] );
+			keys = ch_malloc( 2 * sizeof(struct berval) );
+			keys[0].bv_val = ch_malloc( LDAP_PVT_INTTYPE_CHARS(long) );
+			keys[0].bv_len = snprintf(keys[0].bv_val,
+					LDAP_PVT_INTTYPE_CHARS(long),
+					"%ld", slap_get_time() + pi->smb_can_change);
+			BER_BVZERO( &keys[1] );
 
-                        ml->sml_desc = ad_sambaPwdCanChange;
-                        ml->sml_op = LDAP_MOD_REPLACE;
+			ml->sml_desc = ad_sambaPwdCanChange;
+			ml->sml_op = LDAP_MOD_REPLACE;
 #ifdef SLAP_MOD_INTERNAL
-                        ml->sml_flags = SLAP_MOD_INTERNAL;
+			ml->sml_flags = SLAP_MOD_INTERNAL;
 #endif
-                        ml->sml_values = keys;
-                        ml->sml_nvalues = NULL;
-                }
+			ml->sml_numvals = 1;
+			ml->sml_values = keys;
+			ml->sml_nvalues = NULL;
+		}
 	}
 #endif /* DO_SAMBA */
+
+#ifdef DO_SHADOW
+	/* shadow stuff */
+	if ( SMBK5PWD_DO_SHADOW( pi ) && is_entry_objectclass(e, oc_shadowAccount, 0 ) ) {
+		struct berval *keys;
+
+		ml = ch_malloc(sizeof(Modifications));
+		if (!qpw->rs_modtail) qpw->rs_modtail = &ml->sml_next;
+		ml->sml_next = qpw->rs_mods;
+		qpw->rs_mods = ml;
+
+		keys = ch_malloc( sizeof(struct berval) * 2);
+		BER_BVZERO( &keys[1] );
+		keys[0].bv_val = ch_malloc( LDAP_PVT_INTTYPE_CHARS(long) );
+		keys[0].bv_len = snprintf(keys[0].bv_val,
+			LDAP_PVT_INTTYPE_CHARS(long),
+			"%ld", (long)(slap_get_time() / (60 * 60 * 24)));
+
+		ml->sml_desc = ad_shadowLastChange;
+		ml->sml_op = LDAP_MOD_REPLACE;
+#ifdef SLAP_MOD_INTERNAL
+		ml->sml_flags = SLAP_MOD_INTERNAL;
+#endif
+		ml->sml_numvals = 1;
+		ml->sml_values = keys;
+		ml->sml_nvalues = NULL;
+	}
+#endif /* DO_SHADOW */
+
 	be_entry_release_r( op, e );
+	qpw->rs_new.bv_val[qpw->rs_new.bv_len] = term;
 
 	return SLAP_CB_CONTINUE;
 }
@@ -604,31 +618,34 @@ enum {
 static ConfigDriver smbk5pwd_cf_func;
 
 /*
- * NOTE: uses OID arcs OLcfgOvAt:6 and OLcfgOvOc:6
+ * NOTE: uses OID arcs OLcfgCtAt:1 and OLcfgCtOc:1
  */
 
 static ConfigTable smbk5pwd_cfats[] = {
 	{ "smbk5pwd-enable", "arg",
 		2, 0, 0, ARG_MAGIC|PC_SMB_ENABLE, smbk5pwd_cf_func,
-		"( OLcfgOvAt:6.1 NAME 'olcSmbK5PwdEnable' "
+		"( OLcfgCtAt:1.1 NAME 'olcSmbK5PwdEnable' "
 		"DESC 'Modules to be enabled' "
+		"EQUALITY caseIgnoreMatch "
 		"SYNTAX OMsDirectoryString )", NULL, NULL },
 	{ "smbk5pwd-must-change", "time",
 		2, 2, 0, ARG_MAGIC|ARG_INT|PC_SMB_MUST_CHANGE, smbk5pwd_cf_func,
-		"( OLcfgOvAt:6.2 NAME 'olcSmbK5PwdMustChange' "
+		"( OLcfgCtAt:1.2 NAME 'olcSmbK5PwdMustChange' "
 		"DESC 'Credentials validity interval' "
+		"EQUALITY integerMatch "
+		"SYNTAX OMsInteger SINGLE-VALUE )", NULL, NULL },
+	{ "smbk5pwd-can-change", "time",
+		2, 2, 0, ARG_MAGIC|ARG_INT|PC_SMB_CAN_CHANGE, smbk5pwd_cf_func,
+		"( OLcfgCtAt:1.3 NAME 'olcSmbK5PwdCanChange' "
+		"DESC 'Credentials minimum validity interval' "
+		"EQUALITY integerMatch "
 		"SYNTAX OMsInteger SINGLE-VALUE )", NULL, NULL },
-        { "smbk5pwd-can-change", "time",
-                2, 2, 0, ARG_MAGIC|ARG_INT|PC_SMB_CAN_CHANGE, smbk5pwd_cf_func,
-                "( OLcfgOvAt:6.3 NAME 'olcSmbK5PwdCanChange' "
-                "DESC 'Credentials minimum validity interval' "
-                "SYNTAX OMsInteger SINGLE-VALUE )", NULL, NULL },
 
 	{ NULL, NULL, 0, 0, 0, ARG_IGNORED }
 };
 
 static ConfigOCs smbk5pwd_cfocs[] = {
-	{ "( OLcfgOvOc:6.1 "
+	{ "( OLcfgCtOc:1.1 "
 		"NAME 'olcSmbK5PwdConfig' "
 		"DESC 'smbk5pwd overlay configuration' "
 		"SUP olcOverlayConfig "
@@ -648,6 +665,7 @@ static ConfigOCs smbk5pwd_cfocs[] = {
 static slap_verbmasks smbk5pwd_modules[] = {
 	{ BER_BVC( "krb5" ),		SMBK5PWD_F_KRB5	},
 	{ BER_BVC( "samba" ),		SMBK5PWD_F_SAMBA },
+	{ BER_BVC( "shadow" ),		SMBK5PWD_F_SHADOW },
 	{ BER_BVNULL,			-1 }
 };
 
@@ -669,13 +687,13 @@ smbk5pwd_cf_func( ConfigArgs *c )
 #endif /* ! DO_SAMBA */
 			break;
 
-                case PC_SMB_CAN_CHANGE:
+		case PC_SMB_CAN_CHANGE:
 #ifdef DO_SAMBA
-                        c->value_int = pi->smb_can_change;
+			c->value_int = pi->smb_can_change;
 #else /* ! DO_SAMBA */
-                        c->value_int = 0;
+			c->value_int = 0;
 #endif /* ! DO_SAMBA */
-                        break;
+			break;
 
 		case PC_SMB_ENABLE:
 			c->rvalue_vals = NULL;
@@ -706,10 +724,10 @@ smbk5pwd_cf_func( ConfigArgs *c )
 				pi->mode = 0;
 
 			} else {
-				slap_mask_t	m;
+				int i;
 
-				m = verb_to_mask( c->line, smbk5pwd_modules );
-				pi->mode &= ~m;
+				i = verb_to_mask( c->line, smbk5pwd_modules );
+				pi->mode &= ~smbk5pwd_modules[i].mask;
 			}
 			break;
 
@@ -726,7 +744,7 @@ smbk5pwd_cf_func( ConfigArgs *c )
 		if ( c->value_int < 0 ) {
 			Debug( LDAP_DEBUG_ANY, "%s: smbk5pwd: "
 				"<%s> invalid negative value \"%d\".",
-				c->log, c->argv[ 0 ], 0 );
+				c->log, c->argv[ 0 ], c->value_int );
 			return 1;
 		}
 		pi->smb_must_change = c->value_int;
@@ -734,7 +752,7 @@ smbk5pwd_cf_func( ConfigArgs *c )
 		Debug( LDAP_DEBUG_ANY, "%s: smbk5pwd: "
 			"<%s> only meaningful "
 			"when compiled with -DDO_SAMBA.\n",
-			c->log, c->argv[ 0 ], 0 );
+			c->log, c->argv[ 0 ] );
 		return 1;
 #endif /* ! DO_SAMBA */
 		break;
@@ -744,7 +762,7 @@ smbk5pwd_cf_func( ConfigArgs *c )
                 if ( c->value_int < 0 ) {
                         Debug( LDAP_DEBUG_ANY, "%s: smbk5pwd: "
                                 "<%s> invalid negative value \"%d\".",
-                                c->log, c->argv[ 0 ], 0 );
+                                c->log, c->argv[ 0 ], c->value_int );
                         return 1;
                 }
                 pi->smb_can_change = c->value_int;
@@ -752,13 +770,13 @@ smbk5pwd_cf_func( ConfigArgs *c )
                 Debug( LDAP_DEBUG_ANY, "%s: smbk5pwd: "
                         "<%s> only meaningful "
                         "when compiled with -DDO_SAMBA.\n",
-                        c->log, c->argv[ 0 ], 0 );
+                        c->log, c->argv[ 0 ] );
                 return 1;
 #endif /* ! DO_SAMBA */
                 break;
 
 	case PC_SMB_ENABLE: {
-		slap_mask_t	mode = pi->mode, m;
+		slap_mask_t	mode = pi->mode, m = 0;
 
 		rc = verbs_to_mask( c->argc, c->argv, smbk5pwd_modules, &m );
 		if ( rc > 0 ) {
@@ -793,17 +811,22 @@ smbk5pwd_cf_func( ConfigArgs *c )
 		}
 #endif /* ! DO_SAMBA */
 
-		{
-			BackendDB	db = *c->be;
+#ifndef DO_SHADOW
+		if ( SMBK5PWD_DO_SHADOW( pi ) ) {
+			Debug( LDAP_DEBUG_ANY, "%s: smbk5pwd: "
+				"<%s> module \"%s\" only allowed when compiled with -DDO_SHADOW.\n",
+				c->log, c->argv[ 0 ], c->argv[ rc ] );
+			pi->mode = mode;
+			return 1;
+		}
+#endif /* ! DO_SHADOW */
 
-			/* Re-initialize the module, because
-			 * the configuration might have changed */
-			db.bd_info = (BackendInfo *)on;
-			rc = smbk5pwd_modules_init( pi );
-			if ( rc ) {
-				pi->mode = mode;
-				return 1;
-			}
+		/* Re-initialize the module, because
+		 * the configuration might have changed */
+		rc = smbk5pwd_modules_init( pi );
+		if ( rc ) {
+			pi->mode = mode;
+			return 1;
 		}
 
 		} break;
@@ -827,23 +850,29 @@ smbk5pwd_modules_init( smbk5pwd_t *pi )
 		{ "krb5Key",			&ad_krb5Key },
 		{ "krb5KeyVersionNumber",	&ad_krb5KeyVersionNumber },
 		{ "krb5PrincipalName",		&ad_krb5PrincipalName },
+		{ "krb5ValidEnd",		&ad_krb5ValidEnd },
 		{ NULL }
 	},
 #endif /* DO_KRB5 */
 #ifdef DO_SAMBA
 	samba_ad[] = {
-		{ "sambaLMPassword",		&ad_sambaLMPassword },
 		{ "sambaNTPassword",		&ad_sambaNTPassword },
 		{ "sambaPwdLastSet",		&ad_sambaPwdLastSet },
 		{ "sambaPwdMustChange",		&ad_sambaPwdMustChange },
-                { "sambaPwdCanChange",          &ad_sambaPwdCanChange },
+		{ "sambaPwdCanChange",		&ad_sambaPwdCanChange },
 		{ NULL }
 	},
 #endif /* DO_SAMBA */
+#ifdef DO_SHADOW
+	shadow_ad[] = {
+		{ "shadowLastChange",		&ad_shadowLastChange },
+		{ NULL }
+	},
+#endif /* DO_SHADOW */
 	dummy_ad;
 
 	/* this is to silence the unused var warning */
-	dummy_ad.name = NULL;
+	(void) dummy_ad;
 
 #ifdef DO_KRB5
 	if ( SMBK5PWD_DO_KRB5( pi ) && oc_krb5KDCEntry == NULL ) {
@@ -856,8 +885,7 @@ smbk5pwd_modules_init( smbk5pwd_t *pi )
 		oc_krb5KDCEntry = oc_find( "krb5KDCEntry" );
 		if ( !oc_krb5KDCEntry ) {
 			Debug( LDAP_DEBUG_ANY, "smbk5pwd: "
-				"unable to find \"krb5KDCEntry\" objectClass.\n",
-				0, 0, 0 );
+				"unable to find \"krb5KDCEntry\" objectClass.\n" );
 			return -1;
 		}
 
@@ -881,7 +909,7 @@ smbk5pwd_modules_init( smbk5pwd_t *pi )
 		if (ret) {
 			Debug( LDAP_DEBUG_ANY, "smbk5pwd: "
 				"unable to initialize krb5 context (%d).\n",
-				ret, 0, 0 );
+				ret );
 			oc_krb5KDCEntry = NULL;
 			return -1;
 		}
@@ -895,10 +923,10 @@ smbk5pwd_modules_init( smbk5pwd_t *pi )
 			char *err_str, *err_msg = "<unknown error>";
 			err_str = krb5_get_error_string( context );
 			if (!err_str)
-				err_msg = krb5_get_err_text( context, ret );
+				err_msg = (char *)krb5_get_err_text( context, ret );
 			Debug( LDAP_DEBUG_ANY, "smbk5pwd: "
 				"unable to initialize krb5 admin context: %s (%d).\n",
-				err_str ? err_str : err_msg, ret, 0 );
+				err_str ? err_str : err_msg, ret );
 			if (err_str)
 				krb5_free_error_string( context, err_str );
 			krb5_free_context( context );
@@ -917,8 +945,7 @@ smbk5pwd_modules_init( smbk5pwd_t *pi )
 		oc_sambaSamAccount = oc_find( "sambaSamAccount" );
 		if ( !oc_sambaSamAccount ) {
 			Debug( LDAP_DEBUG_ANY, "smbk5pwd: "
-				"unable to find \"sambaSamAccount\" objectClass.\n",
-				0, 0, 0 );
+				"unable to find \"sambaSamAccount\" objectClass.\n" );
 			return -1;
 		}
 
@@ -939,6 +966,34 @@ smbk5pwd_modules_init( smbk5pwd_t *pi )
 	}
 #endif /* DO_SAMBA */
 
+#ifdef DO_SHADOW
+	if ( SMBK5PWD_DO_SHADOW( pi ) && oc_shadowAccount == NULL ) {
+		int		i, rc;
+
+		oc_shadowAccount = oc_find( "shadowAccount" );
+		if ( !oc_shadowAccount ) {
+			Debug( LDAP_DEBUG_ANY, "smbk5pwd: "
+				"unable to find \"shadowAccount\" objectClass.\n" );
+			return -1;
+		}
+
+		for ( i = 0; shadow_ad[ i ].name != NULL; i++ ) {
+			const char	*text;
+
+			*(shadow_ad[ i ].adp) = NULL;
+
+			rc = slap_str2ad( shadow_ad[ i ].name, shadow_ad[ i ].adp, &text );
+			if ( rc != LDAP_SUCCESS ) {
+				Debug( LDAP_DEBUG_ANY, "smbk5pwd: "
+					"unable to find \"%s\" attributeType: %s (%d).\n",
+					shadow_ad[ i ].name, text, rc );
+				oc_shadowAccount = NULL;
+				return rc;
+			}
+		}
+	}
+#endif /* DO_SHADOW */
+
 	return 0;
 }
 
@@ -997,6 +1052,7 @@ smbk5pwd_initialize(void)
 
 	smbk5pwd.on_bi.bi_type = "smbk5pwd";
 
+	smbk5pwd.on_bi.bi_flags = SLAPO_BFLAG_SINGLE;
 	smbk5pwd.on_bi.bi_db_init = smbk5pwd_db_init;
 	smbk5pwd.on_bi.bi_db_open = smbk5pwd_db_open;
 	smbk5pwd.on_bi.bi_db_destroy = smbk5pwd_db_destroy;
