@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-sql/schema-map.c,v 1.44.2.3 2005/04/29 21:29:10 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2005 The OpenLDAP Foundation.
+ * Copyright 1999-2021 The OpenLDAP Foundation.
  * Portions Copyright 1999 Dmitry Kovalev.
  * Portions Copyright 2002 Pierangelo Masarati.
  * Portions Copyright 2004 Mark Adamson.
@@ -27,6 +27,7 @@
 #include <sys/types.h>
 #include "ac/string.h"
 
+#include "lutil.h"
 #include "slap.h"
 #include "proto-sql.h"
 
@@ -73,6 +74,14 @@ backsql_cmp_attr( const void *v_m1, const void *v_m2 )
 	const backsql_at_map_rec	*m1 = v_m1,
 					*m2 = v_m2;
 
+	if ( slap_ad_is_binary( m1->bam_ad ) || slap_ad_is_binary( m2->bam_ad ) ) {
+#ifdef BACKSQL_USE_PTR_CMP
+		return SLAP_PTRCMP( m1->bam_ad->ad_type, m2->bam_ad->ad_type );
+#else /* ! BACKSQL_USE_PTR_CMP */
+		return ber_bvcmp( &m1->bam_ad->ad_type->sat_cname, &m2->bam_ad->ad_type->sat_cname );
+#endif /* ! BACKSQL_USE_PTR_CMP */
+	}
+
 #ifdef BACKSQL_USE_PTR_CMP
 	return SLAP_PTRCMP( m1->bam_ad, m2->bam_ad );
 #else /* ! BACKSQL_USE_PTR_CMP */
@@ -86,12 +95,26 @@ backsql_dup_attr( void *v_m1, void *v_m2 )
 	backsql_at_map_rec		*m1 = v_m1,
 					*m2 = v_m2;
 
-	assert( m1->bam_ad == m2->bam_ad );
+	if ( slap_ad_is_binary( m1->bam_ad ) || slap_ad_is_binary( m2->bam_ad ) ) {
+#ifdef BACKSQL_USE_PTR_CMP
+		assert( m1->bam_ad->ad_type == m2->bam_ad->ad_type );
+#else /* ! BACKSQL_USE_PTR_CMP */
+		assert( ber_bvcmp( &m1->bam_ad->ad_type->sat_cname, &m2->bam_ad->ad_type->sat_cname ) == 0 );
+#endif /* ! BACKSQL_USE_PTR_CMP */
+
+	} else {
+#ifdef BACKSQL_USE_PTR_CMP
+		assert( m1->bam_ad == m2->bam_ad );
+#else /* ! BACKSQL_USE_PTR_CMP */
+		assert( ber_bvcmp( &m1->bam_ad->ad_cname, &m2->bam_ad->ad_cname ) == 0 );
+#endif /* ! BACKSQL_USE_PTR_CMP */
+	}
 
 	/* duplicate definitions of attributeTypes are appended;
 	 * this allows to define multiple rules for the same 
 	 * attributeType.  Use with care! */
 	for ( ; m1->bam_next ; m1 = m1->bam_next );
+
 	m1->bam_next = m2;
 	m2->bam_next = NULL;
 
@@ -100,19 +123,21 @@ backsql_dup_attr( void *v_m1, void *v_m2 )
 
 static int
 backsql_make_attr_query( 
+	backsql_info		*bi,
 	backsql_oc_map_rec 	*oc_map,
 	backsql_at_map_rec 	*at_map )
 {
 	struct berbuf	bb = BB_NULL;
 
-	backsql_strfcat( &bb, "lblblblbcbl", 
+	backsql_strfcat_x( &bb, NULL, "lblbbbblblbcbl", 
 			(ber_len_t)STRLENOF( "SELECT " ), "SELECT ", 
 			&at_map->bam_sel_expr, 
-			(ber_len_t)STRLENOF( " " BACKSQL_ALIASING BACKSQL_ALIASING_QUOTE ),
-				" " BACKSQL_ALIASING BACKSQL_ALIASING_QUOTE, 
+			(ber_len_t)STRLENOF( " " ), " ",
+			&bi->sql_aliasing,
+			&bi->sql_aliasing_quote, 
 			&at_map->bam_ad->ad_cname,
-			(ber_len_t)STRLENOF( BACKSQL_ALIASING_QUOTE " FROM " ),
-				BACKSQL_ALIASING_QUOTE " FROM ", 
+			&bi->sql_aliasing_quote,
+			(ber_len_t)STRLENOF( " FROM " ), " FROM ", 
 			&at_map->bam_from_tbls, 
 			(ber_len_t)STRLENOF( " WHERE " ), " WHERE ", 
 			&oc_map->bom_keytbl,
@@ -121,25 +146,29 @@ backsql_make_attr_query(
 			(ber_len_t)STRLENOF( "=?" ), "=?" );
 
 	if ( !BER_BVISNULL( &at_map->bam_join_where ) ) {
-		backsql_strfcat( &bb, "lb",
+		backsql_strfcat_x( &bb, NULL, "lb",
 				(ber_len_t)STRLENOF( " AND " ), " AND ", 
 				&at_map->bam_join_where );
 	}
 
-	backsql_strfcat( &bb, "lbl", 
-			(ber_len_t)STRLENOF( " ORDER BY " BACKSQL_ALIASING_QUOTE ),
-				" ORDER BY " BACKSQL_ALIASING_QUOTE,
+	backsql_strfcat_x( &bb, NULL, "lbbb", 
+			(ber_len_t)STRLENOF( " ORDER BY " ), " ORDER BY ",
+			&bi->sql_aliasing_quote,
 			&at_map->bam_ad->ad_cname,
-			(ber_len_t)STRLENOF( BACKSQL_ALIASING_QUOTE ),
-				BACKSQL_ALIASING_QUOTE );
+			&bi->sql_aliasing_quote );
 
 	at_map->bam_query = bb.bb_val.bv_val;
 
 #ifdef BACKSQL_COUNTQUERY
-	/* Query to count how many rows will be returned. */
+	/* Query to count how many rows will be returned.
+
+	SELECT COUNT(*) FROM <from_tbls> WHERE <keytbl>.<keycol>=?
+		[ AND <join_where> ]
+
+	 */
 	BER_BVZERO( &bb.bb_val );
 	bb.bb_len = 0;
-	backsql_strfcat( &bb, "lblbcbl", 
+	backsql_strfcat_x( &bb, NULL, "lblbcbl", 
 			(ber_len_t)STRLENOF( "SELECT COUNT(*) FROM " ),
 				"SELECT COUNT(*) FROM ", 
 			&at_map->bam_from_tbls, 
@@ -150,7 +179,7 @@ backsql_make_attr_query(
 			(ber_len_t)STRLENOF( "=?" ), "=?" );
 
 	if ( !BER_BVISNULL( &at_map->bam_join_where ) ) {
-		backsql_strfcat( &bb, "lb",
+		backsql_strfcat_x( &bb, NULL, "lb",
 				(ber_len_t)STRLENOF( " AND " ), " AND ", 
 				&at_map->bam_join_where );
 	}
@@ -162,20 +191,21 @@ backsql_make_attr_query(
 }
 
 static int
-backsql_add_sysmaps( backsql_oc_map_rec *oc_map )
+backsql_add_sysmaps( backsql_info *bi, backsql_oc_map_rec *oc_map )
 {
 	backsql_at_map_rec	*at_map;
-	char			s[] = "+9223372036854775807L";
+	char			s[LDAP_PVT_INTTYPE_CHARS(long)];
 	struct berval		sbv;
 	struct berbuf		bb;
 	
 	sbv.bv_val = s;
-	sbv.bv_len = snprintf( s, sizeof( s ), "%ld", oc_map->bom_id );
+	sbv.bv_len = snprintf( s, sizeof( s ), BACKSQL_IDNUMFMT, oc_map->bom_id );
 
 	/* extra objectClasses */
 	at_map = (backsql_at_map_rec *)ch_calloc(1, 
 			sizeof( backsql_at_map_rec ) );
 	at_map->bam_ad = slap_schema.si_ad_objectClass;
+	at_map->bam_true_ad = slap_schema.si_ad_objectClass;
 	ber_str2bv( "ldap_entry_objclasses.oc_name", 0, 1,
 			&at_map->bam_sel_expr );
 	ber_str2bv( "ldap_entry_objclasses,ldap_entries", 0, 1, 
@@ -183,12 +213,12 @@ backsql_add_sysmaps( backsql_oc_map_rec *oc_map )
 	
 	bb.bb_len = at_map->bam_from_tbls.bv_len + 1;
 	bb.bb_val = at_map->bam_from_tbls;
-	backsql_merge_from_clause( &bb, &oc_map->bom_keytbl );
+	backsql_merge_from_clause( bi, &bb, &oc_map->bom_keytbl );
 	at_map->bam_from_tbls = bb.bb_val;
 
 	BER_BVZERO( &bb.bb_val );
 	bb.bb_len = 0;
-	backsql_strfcat( &bb, "lbcblb",
+	backsql_strfcat_x( &bb, NULL, "lbcblb",
 			(ber_len_t)STRLENOF( "ldap_entries.id=ldap_entry_objclasses.entry_id AND ldap_entries.keyval=" ),
 				"ldap_entries.id=ldap_entry_objclasses.entry_id AND ldap_entries.keyval=",
 			&oc_map->bom_keytbl, 
@@ -203,34 +233,30 @@ backsql_add_sysmaps( backsql_oc_map_rec *oc_map )
 
 	at_map->bam_add_proc = NULL;
 	{
-		char	tmp[] =
-			"INSERT INTO ldap_entry_objclasses "
+		char	tmp[STRLENOF("INSERT INTO ldap_entry_objclasses "
 			"(entry_id,oc_name) VALUES "
 			"((SELECT id FROM ldap_entries "
-			"WHERE oc_map_id="
-			"18446744073709551615UL "	/* 64 bit ULONG */
-			"AND keyval=?),?)";
+			"WHERE oc_map_id= "
+			"AND keyval=?),?)") + LDAP_PVT_INTTYPE_CHARS(unsigned long)];
 		snprintf( tmp, sizeof(tmp), 
 			"INSERT INTO ldap_entry_objclasses "
 			"(entry_id,oc_name) VALUES "
 			"((SELECT id FROM ldap_entries "
-			"WHERE oc_map_id=%lu "
+			"WHERE oc_map_id=" BACKSQL_IDNUMFMT " "
 			"AND keyval=?),?)", oc_map->bom_id );
 		at_map->bam_add_proc = ch_strdup( tmp );
 	}
 
 	at_map->bam_delete_proc = NULL;
 	{
-		char	tmp[] =
-			"DELETE FROM ldap_entry_objclasses "
+		char	tmp[STRLENOF("DELETE FROM ldap_entry_objclasses "
 			"WHERE entry_id=(SELECT id FROM ldap_entries "
-			"WHERE oc_map_id="
-			"18446744073709551615UL "	/* 64 bit ULONG */
-			"AND keyval=?) AND oc_name=?";
+			"WHERE oc_map_id= "
+			"AND keyval=?) AND oc_name=?") + LDAP_PVT_INTTYPE_CHARS(unsigned long)];
 		snprintf( tmp, sizeof(tmp), 
 			"DELETE FROM ldap_entry_objclasses "
 			"WHERE entry_id=(SELECT id FROM ldap_entries "
-			"WHERE oc_map_id=%lu"
+			"WHERE oc_map_id=" BACKSQL_IDNUMFMT " "
 			"AND keyval=?) AND oc_name=?",
 			oc_map->bom_id );
 		at_map->bam_delete_proc = ch_strdup( tmp );
@@ -240,12 +266,12 @@ backsql_add_sysmaps( backsql_oc_map_rec *oc_map )
 	at_map->bam_expect_return = 0;
 	at_map->bam_next = NULL;
 
-	backsql_make_attr_query( oc_map, at_map );
+	backsql_make_attr_query( bi, oc_map, at_map );
 	if ( avl_insert( &oc_map->bom_attrs, at_map, backsql_cmp_attr, backsql_dup_attr ) == BACKSQL_DUPLICATE ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_add_sysmaps(): "
 				"duplicate attribute \"%s\" in objectClass \"%s\" map\n",
 				at_map->bam_ad->ad_cname.bv_val,
-				oc_map->bom_oc->soc_cname.bv_val, 0 );
+				oc_map->bom_oc->soc_cname.bv_val );
 	}
 
 	/* FIXME: we need to correct the objectClass join_where 
@@ -253,7 +279,7 @@ backsql_add_sysmaps( backsql_oc_map_rec *oc_map )
 	ch_free( at_map->bam_join_where.bv_val );
 	BER_BVZERO( &bb.bb_val );
 	bb.bb_len = 0;
-	backsql_strfcat( &bb, "lbcblb",
+	backsql_strfcat_x( &bb, NULL, "lbcblb",
 			(ber_len_t)STRLENOF( /* "ldap_entries.id=ldap_entry_objclasses.entry_id AND " */ "ldap_entries.keyval=" ),
 				/* "ldap_entries.id=ldap_entry_objclasses.entry_id AND " */ "ldap_entries.keyval=",
 			&oc_map->bom_keytbl, 
@@ -271,7 +297,7 @@ struct backsql_attr_schema_info {
 	backsql_info	*bas_bi;
 	SQLHDBC		bas_dbh;
 	SQLHSTMT	bas_sth;
-	unsigned long	*bas_oc_id;
+	backsql_key_t	*bas_oc_id;
 	int		bas_rc;
 };
 
@@ -291,7 +317,7 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 		"executing at_query\n"
 		"    \"%s\"\n"
 		"    for objectClass \"%s\"\n"
-		"    with param oc_id=\"%lu\"\n",
+		"    with param oc_id=" BACKSQL_IDNUMFMT "\n",
 		bas->bas_bi->sql_at_query,
 		BACKSQL_OC_NAME( oc_map ),
 		*bas->bas_oc_id );
@@ -302,7 +328,7 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 			"error executing at_query\n"
 			"    \"%s\"\n"
 			"    for objectClass \"%s\"\n"
-			"    with param oc_id=\"%lu\"\n",
+			"    with param oc_id=" BACKSQL_IDNUMFMT "\n",
 			bas->bas_bi->sql_at_query,
 			BACKSQL_OC_NAME( oc_map ),
 			*bas->bas_oc_id );
@@ -315,30 +341,43 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 	backsql_BindRowAsStrings( bas->bas_sth, &at_row );
 	for ( ; rc = SQLFetch( bas->bas_sth ), BACKSQL_SUCCESS( rc ); ) {
 		const char	*text = NULL;
-		char		*next = NULL;
 		struct berval	bv;
 		struct berbuf	bb = BB_NULL;
+		AttributeDescription *ad = NULL;
 
-		Debug( LDAP_DEBUG_TRACE, 
-			"attributeType:\n"
-			"\tname=\"%s\"\n"
-			"\tsel_expr=\"%s\"\n"
-			"\tfrom=\"%s\"\n",
-			at_row.cols[ 0 ], at_row.cols[ 1 ],
-			at_row.cols[ 2 ] );
-		Debug( LDAP_DEBUG_TRACE, 
-			"\tjoin_where=\"%s\"\n"
-			"\tadd_proc=\"%s\"\n"
-			"\tdelete_proc=\"%s\"\n",
-			at_row.cols[ 3 ], at_row.cols[ 4 ],
-			at_row.cols[ 5 ]);
-		/* TimesTen */
-		Debug( LDAP_DEBUG_TRACE, "\tsel_expr_u=\"%s\"\n",
-				at_row.cols[ 8 ], 0, 0 );
-		at_map = (backsql_at_map_rec *)ch_calloc( 1,
-				sizeof( backsql_at_map_rec ) );
-		rc = slap_str2ad( at_row.cols[ 0 ], 
-				&at_map->bam_ad, &text );
+		{
+			struct {
+				int idx;
+				char *name;
+			} required[] = {
+				{ 0, "name" },
+				{ 1, "sel_expr" },
+				{ 2, "from" },
+				{ -1, NULL },
+			};
+			int i;
+
+			for ( i = 0; required[ i ].name != NULL; i++ ) {
+				if ( at_row.value_len[ i ] <= 0 ) {
+					Debug( LDAP_DEBUG_ANY,
+						"backsql_oc_get_attr_mapping(): "
+						"required column #%d \"%s\" is empty\n",
+						required[ i ].idx, required[ i ].name );
+					bas->bas_rc = LDAP_OTHER;
+					return BACKSQL_AVL_STOP;
+				}
+			}
+		}
+
+		Debug(LDAP_DEBUG_TRACE,
+		      "attributeType: " "name=\"%s\" " "sel_expr=\"%s\" " "from=\"%s\" " "join_where=\"%s\" " "add_proc=\"%s\" " "delete_proc=\"%s\" " "sel_expr_u=\"%s\"\n",
+		      at_row.cols[0], at_row.cols[1], at_row.cols[2],
+		      at_row.cols[3] ? at_row.cols[3] : "",
+		      at_row.cols[4] ? at_row.cols[4] : "",
+		      at_row.cols[5] ? at_row.cols[5] : "",
+		      at_row.cols[8] ? at_row.cols[8] : "" );
+
+		rc = slap_str2ad( at_row.cols[ 0 ], &ad, &text );
 		if ( rc != LDAP_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_attr_mapping(): "
 				"attribute \"%s\" for objectClass \"%s\" "
@@ -348,9 +387,33 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 			bas->bas_rc = LDAP_CONSTRAINT_VIOLATION;
 			return BACKSQL_AVL_STOP;
 		}
+		at_map = (backsql_at_map_rec *)ch_calloc( 1,
+				sizeof( backsql_at_map_rec ) );
+		at_map->bam_ad = ad;
+		at_map->bam_true_ad = ad;
+		if ( slap_syntax_is_binary( ad->ad_type->sat_syntax )
+			&& !slap_ad_is_binary( ad ) )
+		{
+			char		buf[ SLAP_TEXT_BUFLEN ];
+			struct berval	bv;
+			const char	*text = NULL;
+
+			bv.bv_val = buf;
+			bv.bv_len = snprintf( buf, sizeof( buf ), "%s;binary",
+				ad->ad_cname.bv_val );
+			at_map->bam_true_ad = NULL;
+			bas->bas_rc = slap_bv2ad( &bv, &at_map->bam_true_ad, &text );
+			if ( bas->bas_rc != LDAP_SUCCESS ) {
+				Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_attr_mapping(): "
+					"unable to fetch attribute \"%s\": %s (%d)\n",
+					buf, text, rc );
+				ch_free( at_map );
+				return BACKSQL_AVL_STOP;
+			}
+		}
 
 		ber_str2bv( at_row.cols[ 1 ], 0, 1, &at_map->bam_sel_expr );
-		if ( at_row.value_len[ 8 ] < 0 ) {
+		if ( at_row.value_len[ 8 ] <= 0 ) {
 			BER_BVZERO( &at_map->bam_sel_expr_u );
 
 		} else {
@@ -359,9 +422,9 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 		}
 
 		ber_str2bv( at_row.cols[ 2 ], 0, 0, &bv );
-		backsql_merge_from_clause( &bb, &bv );
+		backsql_merge_from_clause( bas->bas_bi, &bb, &bv );
 		at_map->bam_from_tbls = bb.bb_val;
-		if ( at_row.value_len[ 3 ] < 0 ) {
+		if ( at_row.value_len[ 3 ] <= 0 ) {
 			BER_BVZERO( &at_map->bam_join_where );
 
 		} else {
@@ -370,33 +433,30 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 		}
 		at_map->bam_add_proc = NULL;
 		if ( at_row.value_len[ 4 ] > 0 ) {
-			at_map->bam_add_proc = ch_strdup( at_row.cols[4] );
+			at_map->bam_add_proc = ch_strdup( at_row.cols[ 4 ] );
 		}
 		at_map->bam_delete_proc = NULL;
 		if ( at_row.value_len[ 5 ] > 0 ) {
 			at_map->bam_delete_proc = ch_strdup( at_row.cols[ 5 ] );
 		}
-		at_map->bam_param_order = strtol( at_row.cols[ 6 ], 
-				&next, 0 );
-		if ( next == at_row.cols[ 6 ] || next[0] != '\0' ) {
+		if ( lutil_atoix( &at_map->bam_param_order, at_row.cols[ 6 ], 0 ) != 0 ) {
 			/* error */
 		}
-		at_map->bam_expect_return = strtol( at_row.cols[ 7 ],
-				&next, 0 );
-		if ( next == at_row.cols[ 7 ] || next[0] != '\0' ) {
+		if ( lutil_atoix( &at_map->bam_expect_return, at_row.cols[ 7 ], 0 ) != 0 ) {
 			/* error */
 		}
-		backsql_make_attr_query( oc_map, at_map );
+		backsql_make_attr_query( bas->bas_bi, oc_map, at_map );
 		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_attr_mapping(): "
 			"preconstructed query \"%s\"\n",
-			at_map->bam_query, 0, 0 );
+			at_map->bam_query );
 		at_map->bam_next = NULL;
 		if ( avl_insert( &oc_map->bom_attrs, at_map, backsql_cmp_attr, backsql_dup_attr ) == BACKSQL_DUPLICATE ) {
 			Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_attr_mapping(): "
 					"duplicate attribute \"%s\" "
 					"in objectClass \"%s\" map\n",
 					at_map->bam_ad->ad_cname.bv_val,
-					oc_map->bom_oc->soc_cname.bv_val, 0 );
+					oc_map->bom_oc->soc_cname.bv_val );
+			ch_free( at_map );
 		}
 
 		if ( !BER_BVISNULL( &bas->bas_bi->sql_upper_func ) &&
@@ -404,7 +464,7 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 		{
 			struct berbuf	bb = BB_NULL;
 
-			backsql_strfcat( &bb, "bcbc",
+			backsql_strfcat_x( &bb, NULL, "bcbc",
 					&bas->bas_bi->sql_upper_func,
 					'(' /* ) */ ,
 					&at_map->bam_sel_expr,
@@ -417,9 +477,9 @@ backsql_oc_get_attr_mapping( void *v_oc, void *v_bas )
 
 	Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(\"%s\"): "
 		"autoadding 'objectClass' and 'ref' mappings\n",
-		BACKSQL_OC_NAME( oc_map ), 0, 0 );
+		BACKSQL_OC_NAME( oc_map ) );
 
-	(void)backsql_add_sysmaps( oc_map );
+	(void)backsql_add_sysmaps( bas->bas_bi, oc_map );
 
 	return BACKSQL_AVL_CONTINUE;
 }
@@ -431,11 +491,14 @@ backsql_load_schema_map( backsql_info *bi, SQLHDBC dbh )
 	SQLHSTMT 			sth = SQL_NULL_HSTMT;
 	RETCODE				rc;
 	BACKSQL_ROW_NTS			oc_row;
-	unsigned long			oc_id;
+	backsql_key_t			oc_id;
 	backsql_oc_map_rec		*oc_map;
 	struct backsql_attr_schema_info	bas;
 
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_load_schema_map()\n", 0, 0, 0 );
+	int				delete_proc_idx = 5;
+	int				create_hint_idx = delete_proc_idx + 2;
+
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_load_schema_map()\n" );
 
 	/* 
 	 * TimesTen : See if the ldap_entries.dn_ru field exists in the schema
@@ -447,7 +510,7 @@ backsql_load_schema_map( backsql_info *bi, SQLHDBC dbh )
 			/* Yes, the field exists */
 			bi->sql_flags |= BSQLF_HAS_LDAPINFO_DN_RU;
    			Debug( LDAP_DEBUG_TRACE, "ldapinfo.dn_ru field exists "
-				"in the schema\n", 0, 0, 0 );
+				"in the schema\n" );
 		} else {
 			/* No such field exists */
 			bi->sql_flags &= ~BSQLF_HAS_LDAPINFO_DN_RU;
@@ -457,13 +520,13 @@ backsql_load_schema_map( backsql_info *bi, SQLHDBC dbh )
 	}
 
 	Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): oc_query \"%s\"\n", 
-			bi->sql_oc_query, 0, 0 );
+			bi->sql_oc_query );
 
 	rc = backsql_Prepare( dbh, &sth, bi->sql_oc_query, 0 );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
 			"error preparing oc_query: \"%s\"\n", 
-			bi->sql_oc_query, 0, 0 );
+			bi->sql_oc_query );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 		return LDAP_OTHER;
 	}
@@ -471,63 +534,114 @@ backsql_load_schema_map( backsql_info *bi, SQLHDBC dbh )
 	rc = SQLExecute( sth );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
-			"error executing oc_query: \n", 0, 0, 0 );
+			"error executing oc_query: \n" );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 		return LDAP_OTHER;
 	}
 
 	backsql_BindRowAsStrings( sth, &oc_row );
 	rc = SQLFetch( sth );
+
+	if ( BACKSQL_CREATE_NEEDS_SELECT( bi ) ) {
+		delete_proc_idx++;
+		create_hint_idx++;
+	}
+
 	for ( ; BACKSQL_SUCCESS( rc ); rc = SQLFetch( sth ) ) {
-		int	colnum;
+		{
+			struct {
+				int idx;
+				char *name;
+			} required[] = {
+				{ 0, "id" },
+				{ 1, "name" },
+				{ 2, "keytbl" },
+				{ 3, "keycol" },
+				{ -1, "expect_return" },
+				{ -1, NULL },
+			};
+			int i;
+
+			required[4].idx = delete_proc_idx + 1;
+
+			for ( i = 0; required[ i ].name != NULL; i++ ) {
+				if ( oc_row.value_len[ required[ i ].idx ] <= 0 ) {
+					Debug( LDAP_DEBUG_ANY,
+						"backsql_load_schema_map(): "
+						"required column #%d \"%s\" is empty\n",
+						required[ i ].idx, required[ i ].name );
+					return LDAP_OTHER;
+				}
+			}
+		}
+
+		Debug(LDAP_DEBUG_TRACE,
+		      "objectClass: " "id=\"%s\" " "name=\"%s\" " "keytbl=\"%s\" " "keycol=\"%s\" " "create_proc=\"%s\" " "create_keyval=\"%s\" " "delete_proc=\"%s\" " "expect_return=\"%s\"" "create_hint=\"%s\" \n",
+		      oc_row.cols[0], oc_row.cols[1], oc_row.cols[2],
+		      oc_row.cols[3],
+		      oc_row.cols[4] ? oc_row.cols[4] : "",
+		      (BACKSQL_CREATE_NEEDS_SELECT(bi) && oc_row.cols[5]) ? oc_row.cols[5] : "",
+		      oc_row.cols[delete_proc_idx] ? oc_row.cols[delete_proc_idx] : "",
+		      oc_row.cols[delete_proc_idx + 1],
+		      ((oc_row.ncols > create_hint_idx) && oc_row.cols[create_hint_idx]) ? oc_row.cols[create_hint_idx] : "" );
 
 		oc_map = (backsql_oc_map_rec *)ch_calloc( 1,
 				sizeof( backsql_oc_map_rec ) );
 
-		oc_map->bom_id = strtol( oc_row.cols[ 0 ], NULL, 0 );
+		if ( BACKSQL_STR2ID( &oc_map->bom_id, oc_row.cols[ 0 ], 0 ) != 0 ) {
+			Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
+				"unable to parse id=\"%s\"\n", 
+				oc_row.cols[ 0 ] );
+			ch_free( oc_map );
+			return LDAP_OTHER;
+		}
 
 		oc_map->bom_oc = oc_find( oc_row.cols[ 1 ] );
 		if ( oc_map->bom_oc == NULL ) {
 			Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
 				"objectClass \"%s\" is not defined in schema\n", 
-				oc_row.cols[ 1 ], 0, 0 );
+				oc_row.cols[ 1 ] );
+			ch_free( oc_map );
 			return LDAP_OTHER;	/* undefined objectClass ? */
 		}
 		
 		ber_str2bv( oc_row.cols[ 2 ], 0, 1, &oc_map->bom_keytbl );
 		ber_str2bv( oc_row.cols[ 3 ], 0, 1, &oc_map->bom_keycol );
-		oc_map->bom_create_proc = ( oc_row.value_len[ 4 ] < 0 ) ? NULL 
+		oc_map->bom_create_proc = ( oc_row.value_len[ 4 ] <= 0 ) ? NULL 
 			: ch_strdup( oc_row.cols[ 4 ] );
 
-		colnum = 5;
 		if ( BACKSQL_CREATE_NEEDS_SELECT( bi ) ) {
-			colnum = 6;
-			oc_map->bom_create_keyval = ( oc_row.value_len[ 5 ] < 0 ) 
+			oc_map->bom_create_keyval = ( oc_row.value_len[ 5 ] <= 0 ) 
 				? NULL : ch_strdup( oc_row.cols[ 5 ] );
 		}
-		oc_map->bom_delete_proc = ( oc_row.value_len[ colnum ] < 0 ) ? NULL 
-			: ch_strdup( oc_row.cols[ colnum ] );
-		oc_map->bom_expect_return = strtol( oc_row.cols[ colnum + 1 ], 
-				NULL, 0 );
+		oc_map->bom_delete_proc = ( oc_row.value_len[ delete_proc_idx ] <= 0 ) ? NULL 
+			: ch_strdup( oc_row.cols[ delete_proc_idx ] );
+		if ( lutil_atoix( &oc_map->bom_expect_return, oc_row.cols[ delete_proc_idx + 1 ], 0 ) != 0 ) {
+			Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
+				"unable to parse expect_return=\"%s\" for objectClass \"%s\"\n", 
+				oc_row.cols[ delete_proc_idx + 1 ], oc_row.cols[ 1 ] );
+			ch_free( oc_map );
+			return LDAP_OTHER;
+		}
 
-		colnum += 2;
-		if ( ( oc_row.ncols > colnum ) &&
-				( oc_row.value_len[ colnum ] > 0 ) )
+		if ( ( oc_row.ncols > create_hint_idx ) &&
+				( oc_row.value_len[ create_hint_idx ] > 0 ) )
 		{
 			const char	*text;
 
 			oc_map->bom_create_hint = NULL;
-			rc = slap_str2ad( oc_row.cols[ colnum ],
+			rc = slap_str2ad( oc_row.cols[ create_hint_idx ],
 					&oc_map->bom_create_hint, &text );
 			if ( rc != SQL_SUCCESS ) {
 				Debug( LDAP_DEBUG_TRACE, "load_schema_map(): "
 						"error matching "
 						"AttributeDescription %s "
 						"in create_hint: %s (%d)\n",
-						oc_row.cols[ colnum ],
+						oc_row.cols[ create_hint_idx ],
 						text, rc );
 				backsql_PrintErrors( bi->sql_db_env, dbh,
 						sth, rc );
+				ch_free( oc_map );
 				return LDAP_OTHER;
 			}
 		}
@@ -541,13 +655,14 @@ backsql_load_schema_map( backsql_info *bi, SQLHDBC dbh )
 		if ( avl_insert( &bi->sql_oc_by_oc, oc_map, backsql_cmp_oc, avl_dup_error ) == -1 ) {
 			Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
 					"duplicate objectClass \"%s\" in objectClass map\n",
-					oc_map->bom_oc->soc_cname.bv_val, 0, 0 );
+					oc_map->bom_oc->soc_cname.bv_val );
+			ch_free( oc_map );
 			return LDAP_OTHER;
 		}
 		if ( avl_insert( &bi->sql_oc_by_id, oc_map, backsql_cmp_oc_id, avl_dup_error ) == -1 ) {
 			Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
 					"duplicate objectClass \"%s\" in objectClass by ID map\n",
-					oc_map->bom_oc->soc_cname.bv_val, 0, 0 );
+					oc_map->bom_oc->soc_cname.bv_val );
 			return LDAP_OTHER;
 		}
 		oc_id = oc_map->bom_id;
@@ -557,25 +672,24 @@ backsql_load_schema_map( backsql_info *bi, SQLHDBC dbh )
 			oc_map->bom_keytbl.bv_val, oc_map->bom_keycol.bv_val );
 		if ( oc_map->bom_create_proc ) {
 			Debug( LDAP_DEBUG_TRACE, "    create_proc=\"%s\"\n",
-				oc_map->bom_create_proc, 0, 0 );
+				oc_map->bom_create_proc );
 		}
 		if ( oc_map->bom_create_keyval ) {
 			Debug( LDAP_DEBUG_TRACE, "    create_keyval=\"%s\"\n",
-				oc_map->bom_create_keyval, 0, 0 );
+				oc_map->bom_create_keyval );
 		}
 		if ( oc_map->bom_create_hint ) {
 			Debug( LDAP_DEBUG_TRACE, "    create_hint=\"%s\"\n", 
-				oc_map->bom_create_hint->ad_cname.bv_val,
-				0, 0 );
+				oc_map->bom_create_hint->ad_cname.bv_val );
 		}
 		if ( oc_map->bom_delete_proc ) {
 			Debug( LDAP_DEBUG_TRACE, "    delete_proc=\"%s\"\n", 
-				oc_map->bom_delete_proc, 0, 0 );
+				oc_map->bom_delete_proc );
 		}
 		Debug( LDAP_DEBUG_TRACE, "    expect_return: "
 			"add=%d, del=%d; attributes:\n",
 			BACKSQL_IS_ADD( oc_map->bom_expect_return ), 
-			BACKSQL_IS_DEL( oc_map->bom_expect_return ), 0 );
+			BACKSQL_IS_DEL( oc_map->bom_expect_return ) );
 	}
 
 	backsql_FreeRow( &oc_row );
@@ -583,21 +697,21 @@ backsql_load_schema_map( backsql_info *bi, SQLHDBC dbh )
 
 	/* prepare for attribute fetching */
 	Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): at_query \"%s\"\n", 
-			bi->sql_at_query, 0, 0 );
+			bi->sql_at_query );
 
 	rc = backsql_Prepare( dbh, &sth, bi->sql_at_query, 0 );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
 			"error preparing at_query: \"%s\"\n", 
-			bi->sql_at_query, 0, 0 );
+			bi->sql_at_query );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 		return LDAP_OTHER;
 	}
 
-	rc = backsql_BindParamInt( sth, 1, SQL_PARAM_INPUT, &oc_id );
+	rc = backsql_BindParamNumID( sth, 1, SQL_PARAM_INPUT, &oc_id );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_load_schema_map(): "
-			"error binding param \"oc_id\" for at_query\n", 0, 0, 0 );
+			"error binding param \"oc_id\" for at_query\n" );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
 		return LDAP_OTHER;
@@ -616,7 +730,7 @@ backsql_load_schema_map( backsql_info *bi, SQLHDBC dbh )
 
 	bi->sql_flags |= BSQLF_SCHEMA_LOADED;
 
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_load_schema_map()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_load_schema_map()\n" );
 
 	return bas.bas_rc;
 }
@@ -629,7 +743,7 @@ backsql_oc2oc( backsql_info *bi, ObjectClass *oc )
 #ifdef BACKSQL_TRACE
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_oc2oc(): "
 		"searching for objectclass with name=\"%s\"\n",
-		oc->soc_cname.bv_val, 0, 0 );
+		oc->soc_cname.bv_val );
 #endif /* BACKSQL_TRACE */
 
 	tmp.bom_oc = oc;
@@ -638,10 +752,10 @@ backsql_oc2oc( backsql_info *bi, ObjectClass *oc )
 	if ( res != NULL ) {
 		Debug( LDAP_DEBUG_TRACE, "<==backsql_oc2oc(): "
 			"found name=\"%s\", id=%d\n", 
-			BACKSQL_OC_NAME( res ), res->bom_id, 0 );
+			BACKSQL_OC_NAME( res ), res->bom_id );
 	} else {
 		Debug( LDAP_DEBUG_TRACE, "<==backsql_oc2oc(): "
-			"not found\n", 0, 0, 0 );
+			"not found\n" );
 	}
 #endif /* BACKSQL_TRACE */
  
@@ -656,7 +770,7 @@ backsql_name2oc( backsql_info *bi, struct berval *oc_name )
 #ifdef BACKSQL_TRACE
 	Debug( LDAP_DEBUG_TRACE, "==>oc_with_name(): "
 		"searching for objectclass with name=\"%s\"\n",
-		oc_name->bv_val, 0, 0 );
+		oc_name->bv_val );
 #endif /* BACKSQL_TRACE */
 
 	tmp.bom_oc = oc_bvfind( oc_name );
@@ -669,10 +783,10 @@ backsql_name2oc( backsql_info *bi, struct berval *oc_name )
 	if ( res != NULL ) {
 		Debug( LDAP_DEBUG_TRACE, "<==oc_with_name(): "
 			"found name=\"%s\", id=%d\n", 
-			BACKSQL_OC_NAME( res ), res->bom_id, 0 );
+			BACKSQL_OC_NAME( res ), res->bom_id );
 	} else {
 		Debug( LDAP_DEBUG_TRACE, "<==oc_with_name(): "
-			"not found\n", 0, 0, 0 );
+			"not found\n" );
 	}
 #endif /* BACKSQL_TRACE */
  
@@ -686,7 +800,7 @@ backsql_id2oc( backsql_info *bi, unsigned long id )
  
 #ifdef BACKSQL_TRACE
 	Debug( LDAP_DEBUG_TRACE, "==>oc_with_id(): "
-		"searching for objectclass with id='%d'\n", id, 0, 0 );
+		"searching for objectclass with id=%lu\n", id );
 #endif /* BACKSQL_TRACE */
 
 	tmp.bom_id = id;
@@ -695,12 +809,12 @@ backsql_id2oc( backsql_info *bi, unsigned long id )
 
 #ifdef BACKSQL_TRACE
 	if ( res != NULL ) {
-		Debug( LDAP_DEBUG_TRACE, "<==oc_with_name(): "
-			"found name=\"%s\", id=%d\n",
-			BACKSQL_OC_NAME( res ), res->bom_id, 0 );
+		Debug( LDAP_DEBUG_TRACE, "<==oc_with_id(): "
+			"found name=\"%s\", id=%lu\n",
+			BACKSQL_OC_NAME( res ), res->bom_id );
 	} else {
-		Debug( LDAP_DEBUG_TRACE, "<==oc_with_name(): "
-			"not found\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_TRACE, "<==oc_with_id(): "
+			"id=%lu not found\n", res->bom_id );
 	}
 #endif /* BACKSQL_TRACE */
 	
@@ -710,12 +824,12 @@ backsql_id2oc( backsql_info *bi, unsigned long id )
 backsql_at_map_rec *
 backsql_ad2at( backsql_oc_map_rec* objclass, AttributeDescription *ad )
 {
-	backsql_at_map_rec	tmp, *res;
+	backsql_at_map_rec	tmp = { 0 }, *res;
  
 #ifdef BACKSQL_TRACE
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_ad2at(): "
 		"searching for attribute \"%s\" for objectclass \"%s\"\n",
-		ad->ad_cname.bv_val, BACKSQL_OC_NAME( objclass ), 0 );
+		ad->ad_cname.bv_val, BACKSQL_OC_NAME( objclass ) );
 #endif /* BACKSQL_TRACE */
 
 	tmp.bam_ad = ad;
@@ -727,10 +841,10 @@ backsql_ad2at( backsql_oc_map_rec* objclass, AttributeDescription *ad )
 		Debug( LDAP_DEBUG_TRACE, "<==backsql_ad2at(): "
 			"found name=\"%s\", sel_expr=\"%s\"\n",
 			res->bam_ad->ad_cname.bv_val,
-			res->bam_sel_expr.bv_val, 0 );
+			res->bam_sel_expr.bv_val );
 	} else {
 		Debug( LDAP_DEBUG_TRACE, "<==backsql_ad2at(): "
-			"not found\n", 0, 0, 0 );
+			"not found\n" );
 	}
 #endif /* BACKSQL_TRACE */
 
@@ -799,9 +913,9 @@ backsql_supad2at( backsql_oc_map_rec *objclass, AttributeDescription *supad,
 	struct supad2at_t	va = { 0 };
 	int			rc;
 
-	assert( objclass );
-	assert( supad );
-	assert( pret );
+	assert( objclass != NULL );
+	assert( supad != NULL );
+	assert( pret != NULL );
 
 	*pret = NULL;
 
@@ -824,7 +938,7 @@ backsql_free_attr( void *v_at )
 	backsql_at_map_rec	*at = v_at;
 	
 	Debug( LDAP_DEBUG_TRACE, "==>free_attr(): \"%s\"\n", 
-			at->bam_ad->ad_cname.bv_val, 0, 0 );
+			at->bam_ad->ad_cname.bv_val );
 	ch_free( at->bam_sel_expr.bv_val );
 	if ( !BER_BVISNULL( &at->bam_from_tbls ) ) {
 		ch_free( at->bam_from_tbls.bv_val );
@@ -842,6 +956,12 @@ backsql_free_attr( void *v_at )
 		ch_free( at->bam_query );
 	}
 
+#ifdef BACKSQL_COUNTQUERY
+	if ( at->bam_countquery != NULL ) {
+		ch_free( at->bam_countquery );
+	}
+#endif /* BACKSQL_COUNTQUERY */
+
 	/* TimesTen */
 	if ( !BER_BVISNULL( &at->bam_sel_expr_u ) ) {
 		ch_free( at->bam_sel_expr_u.bv_val );
@@ -853,7 +973,7 @@ backsql_free_attr( void *v_at )
 	
 	ch_free( at );
 
-	Debug( LDAP_DEBUG_TRACE, "<==free_attr()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==free_attr()\n" );
 }
 
 static void
@@ -862,7 +982,7 @@ backsql_free_oc( void *v_oc )
 	backsql_oc_map_rec	*oc = v_oc;
 	
 	Debug( LDAP_DEBUG_TRACE, "==>free_oc(): \"%s\"\n", 
-			BACKSQL_OC_NAME( oc ), 0, 0 );
+			BACKSQL_OC_NAME( oc ) );
 	avl_free( oc->bom_attrs, backsql_free_attr );
 	ch_free( oc->bom_keytbl.bv_val );
 	ch_free( oc->bom_keycol.bv_val );
@@ -877,16 +997,16 @@ backsql_free_oc( void *v_oc )
 	}
 	ch_free( oc );
 
-	Debug( LDAP_DEBUG_TRACE, "<==free_oc()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==free_oc()\n" );
 }
 
 int
 backsql_destroy_schema_map( backsql_info *bi )
 {
-	Debug( LDAP_DEBUG_TRACE, "==>destroy_schema_map()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>destroy_schema_map()\n" );
 	avl_free( bi->sql_oc_by_oc, 0 );
 	avl_free( bi->sql_oc_by_id, backsql_free_oc );
-	Debug( LDAP_DEBUG_TRACE, "<==destroy_schema_map()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==destroy_schema_map()\n" );
 	return 0;
 }
 
