@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-sql/search.c,v 1.117.2.5 2007/08/31 23:14:05 quanah Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2007 The OpenLDAP Foundation.
+ * Copyright 1999-2021 The OpenLDAP Foundation.
  * Portions Copyright 1999 Dmitry Kovalev.
  * Portions Copyright 2002 Pierangelo Masarati.
  * Portions Copyright 2004 Mark Adamson.
@@ -48,6 +48,9 @@ static int backsql_process_filter_attr( backsql_srch_info *bsi, Filter *f,
    and the other 26 for ldap_entries ID number. If your max(oc_map_id) is more
    than 63, you will need to steal more bits from ldap_entries ID number and
    put them into the OC ID part of the cookie. */
+
+/* NOTE: not supported when BACKSQL_ARBITRARY_KEY is defined */
+#ifndef BACKSQL_ARBITRARY_KEY
 #define SQL_TO_PAGECOOKIE(id, oc) (((id) << 6 ) | ((oc) & 0x3F))
 #define PAGECOOKIE_TO_SQL_ID(pc) ((pc) >> 6)
 #define PAGECOOKIE_TO_SQL_OC(pc) ((pc) & 0x3F)
@@ -58,6 +61,7 @@ static void send_paged_response(
 	Operation *op,
 	SlapReply *rs,
 	ID  *lastid );
+#endif /* ! BACKSQL_ARBITRARY_KEY */
 
 static int
 backsql_attrlist_add( backsql_srch_info *bsi, AttributeDescription *ad )
@@ -89,7 +93,7 @@ backsql_attrlist_add( backsql_srch_info *bsi, AttributeDescription *ad )
 		
 		Debug( LDAP_DEBUG_TRACE, "==>backsql_attrlist_add(): "
 			"attribute \"%s\" is in list\n", 
-			an->an_name.bv_val, 0, 0 );
+			an->an_name.bv_val );
 		/*
 		 * We can live with strcmp because the attribute 
 		 * list has been normalized before calling be_search
@@ -100,7 +104,7 @@ backsql_attrlist_add( backsql_srch_info *bsi, AttributeDescription *ad )
 	}
 	
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_attrlist_add(): "
-		"adding \"%s\" to list\n", ad->ad_cname.bv_val, 0, 0 );
+		"adding \"%s\" to list\n", ad->ad_cname.bv_val );
 
 	an = (AttributeName *)bsi->bsi_op->o_tmprealloc( bsi->bsi_attrs,
 			sizeof( AttributeName ) * ( n_attrs + 2 ),
@@ -184,7 +188,7 @@ backsql_init_search(
 			BER_BVZERO( &bsi->bsi_attrs[ 0 ].an_name );
 	
 			for ( p = attrs; !BER_BVISNULL( &p->an_name ); p++ ) {
-				if ( BACKSQL_NCMP( &p->an_name, &AllUser ) == 0 ) {
+				if ( BACKSQL_NCMP( &p->an_name, slap_bv_all_user_attrs ) == 0 ) {
 					/* handle "*" */
 					bsi->bsi_flags |= BSQL_SF_ALL_USER;
 
@@ -198,7 +202,7 @@ backsql_init_search(
 					}
 					continue;
 
-				} else if ( BACKSQL_NCMP( &p->an_name, &AllOper ) == 0 ) {
+				} else if ( BACKSQL_NCMP( &p->an_name, slap_bv_all_operational_attrs ) == 0 ) {
 					/* handle "+" */
 					bsi->bsi_flags |= BSQL_SF_ALL_OPER;
 
@@ -212,7 +216,7 @@ backsql_init_search(
 					}
 					continue;
 
-				} else if ( BACKSQL_NCMP( &p->an_name, &NoAttrs ) == 0 ) {
+				} else if ( BACKSQL_NCMP( &p->an_name, slap_bv_no_attrs ) == 0 ) {
 					/* ignore "1.1" */
 					continue;
 
@@ -237,7 +241,7 @@ backsql_init_search(
 			
 			/* use hints if available */
 			for ( p = bi->sql_anlist; !BER_BVISNULL( &p->an_name ); p++ ) {
-				if ( BACKSQL_NCMP( &p->an_name, &AllUser ) == 0 ) {
+				if ( BACKSQL_NCMP( &p->an_name, slap_bv_all_user_attrs ) == 0 ) {
 					/* handle "*" */
 					bsi->bsi_flags |= BSQL_SF_ALL_USER;
 
@@ -251,7 +255,7 @@ backsql_init_search(
 					}
 					continue;
 
-				} else if ( BACKSQL_NCMP( &p->an_name, &AllOper ) == 0 ) {
+				} else if ( BACKSQL_NCMP( &p->an_name, slap_bv_all_operational_attrs ) == 0 ) {
 					/* handle "+" */
 					bsi->bsi_flags |= BSQL_SF_ALL_OPER;
 
@@ -523,7 +527,7 @@ backsql_process_sub_filter( backsql_srch_info *bsi, Filter *f,
 
 	/* TimesTen */
 	Debug( LDAP_DEBUG_TRACE, "backsql_process_sub_filter(%s):\n",
-		at->bam_ad->ad_cname.bv_val, 0, 0 );
+		at->bam_ad->ad_cname.bv_val );
 	Debug(LDAP_DEBUG_TRACE, "   expr: '%s%s%s'\n", at->bam_sel_expr.bv_val,
 		at->bam_sel_expr_u.bv_val ? "' '" : "",
 		at->bam_sel_expr_u.bv_val ? at->bam_sel_expr_u.bv_val : "" );
@@ -554,7 +558,7 @@ backsql_process_sub_filter( backsql_srch_info *bsi, Filter *f,
 		Debug( LDAP_DEBUG_TRACE, 
 			"==>backsql_process_sub_filter(%s): "
 			"sub_initial=\"%s\"\n", at->bam_ad->ad_cname.bv_val,
-			f->f_sub_initial.bv_val, 0 );
+			f->f_sub_initial.bv_val );
 #endif /* BACKSQL_TRACE */
 
 		start = bsi->bsi_flt_where.bb_val.bv_len;
@@ -604,7 +608,7 @@ backsql_process_sub_filter( backsql_srch_info *bsi, Filter *f,
 		Debug( LDAP_DEBUG_TRACE, 
 			"==>backsql_process_sub_filter(%s): "
 			"sub_final=\"%s\"\n", at->bam_ad->ad_cname.bv_val,
-			f->f_sub_final.bv_val, 0 );
+			f->f_sub_final.bv_val );
 #endif /* BACKSQL_TRACE */
 
 		start = bsi->bsi_flt_where.bb_val.bv_len;
@@ -683,7 +687,7 @@ backsql_process_filter( backsql_srch_info *bsi, Filter *f )
 	int 			done = 0;
 	int			rc = 0;
 
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_process_filter()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_process_filter()\n" );
 	if ( f->f_choice == SLAPD_FILTER_COMPUTED ) {
 		struct berval	flt;
 		char		*msg = NULL;
@@ -710,13 +714,23 @@ backsql_process_filter( backsql_srch_info *bsi, Filter *f )
 		}
 
 		Debug( LDAP_DEBUG_TRACE, "backsql_process_filter(): "
-			"filter computed (%s)\n", msg, 0, 0 );
+			"filter computed (%s)\n", msg );
 		backsql_strfcat_x( &bsi->bsi_flt_where,
 				bsi->bsi_op->o_tmpmemctx, "b", &flt );
 		rc = 1;
 		goto done;
 	}
 
+	if ( f->f_choice & SLAPD_FILTER_UNDEFINED ) {
+		backsql_strfcat_x( &bsi->bsi_flt_where,
+			bsi->bsi_op->o_tmpmemctx,
+			"l",
+			(ber_len_t)STRLENOF( "1=0" ), "1=0" );
+		done = 1;
+		rc = 1;
+		goto done;
+	}
+
 	switch( f->f_choice ) {
 	case LDAP_FILTER_OR:
 		rc = backsql_process_filter_list( bsi, f->f_or, 
@@ -803,7 +817,7 @@ backsql_process_filter( backsql_srch_info *bsi, Filter *f )
 						"backsql_process_filter(): "
 						"unknown objectClass \"%s\" "
 						"in filter\n",
-						f->f_av_value.bv_val, 0, 0 );
+						f->f_av_value.bv_val );
 				bsi->bsi_status = LDAP_OTHER;
 				rc = -1;
 				goto done;
@@ -851,8 +865,7 @@ backsql_process_filter( backsql_srch_info *bsi, Filter *f )
 			Debug( LDAP_DEBUG_TRACE,
 					"backsql_process_filter(): "
 					"illegal/unhandled filter "
-					"on objectClass attribute",
-					0, 0, 0 );
+					"on objectClass attribute" );
 			bsi->bsi_status = LDAP_OTHER;
 			rc = -1;
 			goto done;
@@ -864,7 +877,7 @@ backsql_process_filter( backsql_srch_info *bsi, Filter *f )
 		struct berval	keyval;
 #else /* ! BACKSQL_ARBITRARY_KEY */
 		unsigned long	keyval;
-		char		keyvalbuf[] = "18446744073709551615";
+		char		keyvalbuf[LDAP_PVT_INTTYPE_CHARS(unsigned long)];
 #endif /* ! BACKSQL_ARBITRARY_KEY */
 
 		switch ( f->f_choice ) {
@@ -914,7 +927,7 @@ backsql_process_filter( backsql_srch_info *bsi, Filter *f )
 #ifdef BACKSQL_SYNCPROV
 	} else if ( ad == slap_schema.si_ad_entryCSN ) {
 		/*
-		 * support for syncrepl as producer...
+		 * support for syncrepl as provider...
 		 */
 #if 0
 		if ( !bsi->bsi_op->o_sync ) {
@@ -999,7 +1012,7 @@ backsql_process_filter( backsql_srch_info *bsi, Filter *f )
 
 	if ( vat == NULL ) {
 		/* search anyway; other parts of the filter
-		 * may succeeed */
+		 * may succeed */
 		backsql_strfcat_x( &bsi->bsi_flt_where,
 				bsi->bsi_op->o_tmpmemctx,
 				"l",
@@ -1061,7 +1074,7 @@ done:;
 
 	Debug( LDAP_DEBUG_TRACE,
 			"<==backsql_process_filter() %s\n",
-			rc == 1 ? "succeeded" : "failed", 0, 0);
+			rc == 1 ? "succeeded" : "failed" );
 
 	return rc;
 }
@@ -1171,7 +1184,7 @@ backsql_process_filter_attr( backsql_srch_info *bsi, Filter *f, backsql_at_map_r
 	struct berval		ordering = BER_BVC("<=");
 
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_process_filter_attr(%s)\n",
-		at->bam_ad->ad_cname.bv_val, 0, 0 );
+		at->bam_ad->ad_cname.bv_val );
 
 	/*
 	 * need to add this attribute to list of attrs to load,
@@ -1192,6 +1205,14 @@ backsql_process_filter_attr( backsql_srch_info *bsi, Filter *f, backsql_at_map_r
 				&at->bam_join_where );
 	}
 
+	if ( f->f_choice & SLAPD_FILTER_UNDEFINED ) {
+		backsql_strfcat_x( &bsi->bsi_flt_where,
+			bsi->bsi_op->o_tmpmemctx,
+			"l",
+			(ber_len_t)STRLENOF( "1=0" ), "1=0" );
+		return 1;
+	}
+
 	switch ( f->f_choice ) {
 	case LDAP_FILTER_EQUALITY:
 		filter_value = &f->f_av_value;
@@ -1373,7 +1394,7 @@ equality_match:;
 	}
 
 	Debug( LDAP_DEBUG_TRACE, "<==backsql_process_filter_attr(%s)\n",
-		at->bam_ad->ad_cname.bv_val, 0, 0 );
+		at->bam_ad->ad_cname.bv_val );
 
 	return 1;
 }
@@ -1389,7 +1410,7 @@ backsql_srch_query( backsql_srch_info *bsi, struct berval *query )
 
 	bsi->bsi_use_subtree_shortcut = 0;
 
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_srch_query()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_srch_query()\n" );
 	BER_BVZERO( &bsi->bsi_sel.bb_val );
 	BER_BVZERO( &bsi->bsi_sel.bb_val );
 	bsi->bsi_sel.bb_len = 0;
@@ -1551,6 +1572,7 @@ backsql_srch_query( backsql_srch_info *bsi, struct berval *query )
 		assert( 0 );
 	}
 
+#ifndef BACKSQL_ARBITRARY_KEY
 	/* If paged results are in effect, ignore low ldap_entries.id numbers */
 	if ( get_pagedresults(bsi->bsi_op) > SLAP_CONTROL_IGNORED ) {
 		unsigned long lowid = 0;
@@ -1574,6 +1596,7 @@ backsql_srch_query( backsql_srch_info *bsi, struct berval *query )
 					lowidstring );
 		}
 	}
+#endif /* ! BACKSQL_ARBITRARY_KEY */
 
 	rc = backsql_process_filter( bsi, bsi->bsi_filter );
 	if ( rc > 0 ) {
@@ -1613,7 +1636,7 @@ backsql_srch_query( backsql_srch_info *bsi, struct berval *query )
 	bsi->bsi_flt_where.bb_len = 0;
 	
 	Debug( LDAP_DEBUG_TRACE, "<==backsql_srch_query() returns %s\n",
-		query->bv_val ? query->bv_val : "NULL", 0, 0 );
+		query->bv_val ? query->bv_val : "NULL" );
 	
 	return ( rc <= 0 ? 1 : 0 );
 }
@@ -1646,7 +1669,7 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
 	bsi->bsi_status = LDAP_SUCCESS;
  
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_oc_get_candidates(): oc=\"%s\"\n",
-			BACKSQL_OC_NAME( oc ), 0, 0 );
+			BACKSQL_OC_NAME( oc ) );
 
 	/* check for abandon */
 	if ( op->o_abandon ) {
@@ -1654,6 +1677,7 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
 		return BACKSQL_AVL_STOP;
 	}
 
+#ifndef BACKSQL_ARBITRARY_KEY
 	/* If paged results have already completed this objectClass, skip it */
 	if ( get_pagedresults(op) > SLAP_CONTROL_IGNORED ) {
 		if ( oc->bom_id < PAGECOOKIE_TO_SQL_OC( ((PagedResultsState *)op->o_pagedresults_state)->ps_cookie ) )
@@ -1661,10 +1685,11 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
 			return BACKSQL_AVL_CONTINUE;
 		}
 	}
+#endif /* ! BACKSQL_ARBITRARY_KEY */
 
 	if ( bsi->bsi_n_candidates == -1 ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
-			"unchecked limit has been overcome\n", 0, 0, 0 );
+			"unchecked limit has been overcome\n" );
 		/* should never get here */
 		assert( 0 );
 		bsi->bsi_status = LDAP_ADMINLIMIT_EXCEEDED;
@@ -1676,7 +1701,7 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
 	if ( res ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
 			"error while constructing query for objectclass \"%s\"\n",
-			oc->bom_oc->soc_cname.bv_val, 0, 0 );
+			oc->bom_oc->soc_cname.bv_val );
 		/*
 		 * FIXME: need to separate errors from legally
 		 * impossible filters
@@ -1701,32 +1726,33 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
 	if ( BER_BVISNULL( &query ) ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
 			"could not construct query for objectclass \"%s\"\n",
-			oc->bom_oc->soc_cname.bv_val, 0, 0 );
+			oc->bom_oc->soc_cname.bv_val );
 		bsi->bsi_status = LDAP_SUCCESS;
 		return BACKSQL_AVL_CONTINUE;
 	}
 
 	Debug( LDAP_DEBUG_TRACE, "Constructed query: %s\n", 
-			query.bv_val, 0, 0 );
+			query.bv_val );
 
 	rc = backsql_Prepare( bsi->bsi_dbh, &sth, query.bv_val, 0 );
 	bsi->bsi_op->o_tmpfree( query.bv_val, bsi->bsi_op->o_tmpmemctx );
 	BER_BVZERO( &query );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
-			"error preparing query\n", 0, 0, 0 );
+			"error preparing query\n" );
 		backsql_PrintErrors( bi->sql_db_env, bsi->bsi_dbh, sth, rc );
 		bsi->bsi_status = LDAP_OTHER;
 		return BACKSQL_AVL_CONTINUE;
 	}
 	
-	Debug( LDAP_DEBUG_TRACE, "id: '%ld'\n", bsi->bsi_oc->bom_id, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "id: '" BACKSQL_IDNUMFMT "'\n",
+		bsi->bsi_oc->bom_id );
 
-	rc = backsql_BindParamInt( sth, 1, SQL_PARAM_INPUT,
+	rc = backsql_BindParamNumID( sth, 1, SQL_PARAM_INPUT,
 			&bsi->bsi_oc->bom_id );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
-			"error binding objectclass id parameter\n", 0, 0, 0 );
+			"error binding objectclass id parameter\n" );
 		bsi->bsi_status = LDAP_OTHER;
 		return BACKSQL_AVL_CONTINUE;
 	}
@@ -1753,13 +1779,13 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
 		}
 
 		Debug( LDAP_DEBUG_TRACE, "(base)dn: \"%s\"\n",
-				tmp_base_ndn, 0, 0 );
+				tmp_base_ndn );
 
 		rc = backsql_BindParamStr( sth, 2, SQL_PARAM_INPUT,
 				tmp_base_ndn, BACKSQL_MAX_DN_LEN );
 		if ( rc != SQL_SUCCESS ) {
          		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
-				"error binding base_ndn parameter\n", 0, 0, 0 );
+				"error binding base_ndn parameter\n" );
 			backsql_PrintErrors( bi->sql_db_env, bsi->bsi_dbh, 
 					sth, rc );
 			bsi->bsi_status = LDAP_OTHER;
@@ -1834,18 +1860,17 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
 
 		if ( bsi->bsi_scope == LDAP_SCOPE_SUBORDINATE ) {
 			Debug( LDAP_DEBUG_TRACE, "(children)dn: \"%s\"\n",
-				tmp_base_ndn, 0, 0 );
+				tmp_base_ndn );
 		} else {
 			Debug( LDAP_DEBUG_TRACE, "(sub)dn: \"%s\"\n",
-				tmp_base_ndn, 0, 0 );
+				tmp_base_ndn );
 		}
 
 		rc = backsql_BindParamStr( sth, 2, SQL_PARAM_INPUT,
 				tmp_base_ndn, BACKSQL_MAX_DN_LEN );
 		if ( rc != SQL_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
-				"error binding base_ndn parameter (2)\n",
-				0, 0, 0 );
+				"error binding base_ndn parameter (2)\n" );
 			backsql_PrintErrors( bi->sql_db_env, bsi->bsi_dbh, 
 					sth, rc );
 			bsi->bsi_status = LDAP_OTHER;
@@ -1857,18 +1882,13 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
  	case LDAP_SCOPE_ONELEVEL:
 		assert( !BER_BVISNULL( &bsi->bsi_base_id.eid_ndn ) );
 
-#ifdef BACKSQL_ARBITRARY_KEY
-		Debug( LDAP_DEBUG_TRACE, "(one)id: \"%s\"\n",
-				bsi->bsi_base_id.eid_id.bv_val, 0, 0 );
-#else /* ! BACKSQL_ARBITRARY_KEY */
-		Debug( LDAP_DEBUG_TRACE, "(one)id: '%lu'\n",
-				bsi->bsi_base_id.eid_id, 0, 0 );
-#endif /* ! BACKSQL_ARBITRARY_KEY */
+		Debug( LDAP_DEBUG_TRACE, "(one)id=" BACKSQL_IDFMT "\n",
+			BACKSQL_IDARG(bsi->bsi_base_id.eid_id) );
 		rc = backsql_BindParamID( sth, 2, SQL_PARAM_INPUT,
 				&bsi->bsi_base_id.eid_id );
 		if ( rc != SQL_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
-				"error binding base id parameter\n", 0, 0, 0 );
+				"error binding base id parameter\n" );
 			bsi->bsi_status = LDAP_OTHER;
 			return BACKSQL_AVL_CONTINUE;
 		}
@@ -1878,7 +1898,7 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
 	rc = SQLExecute( sth );
 	if ( !BACKSQL_SUCCESS( rc ) ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
-			"error executing query\n", 0, 0, 0 );
+			"error executing query\n" );
 		backsql_PrintErrors( bi->sql_db_env, bsi->bsi_dbh, sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
 		bsi->bsi_status = LDAP_OTHER;
@@ -1919,10 +1939,10 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
 		ber_str2bv_x( row.cols[ 1 ], 0, 1, &c_id->eid_keyval,
 				op->o_tmpmemctx );
 #else /* ! BACKSQL_ARBITRARY_KEY */
-		if ( lutil_atoulx( &c_id->eid_id, row.cols[ 0 ], 0 ) != 0 ) {
+		if ( BACKSQL_STR2ID( &c_id->eid_id, row.cols[ 0 ], 0 ) != 0 ) {
 			goto cleanup;
 		}
-		if ( lutil_atoulx( &c_id->eid_keyval, row.cols[ 1 ], 0 ) != 0 ) {
+		if ( BACKSQL_STR2ID( &c_id->eid_keyval, row.cols[ 1 ], 0 ) != 0 ) {
 			goto cleanup;
 		}
 #endif /* ! BACKSQL_ARBITRARY_KEY */
@@ -1937,16 +1957,11 @@ backsql_oc_get_candidates( void *v_oc, void *v_bsi )
 		*bsi->bsi_id_listtail = c_id;
 		bsi->bsi_id_listtail = &c_id->eid_next;
 
-#ifdef BACKSQL_ARBITRARY_KEY
 		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
-			"added entry id=%s, keyval=%s dn=\"%s\"\n",
-			c_id->eid_id.bv_val, c_id->eid_keyval.bv_val,
+			"added entry id=" BACKSQL_IDFMT " keyval=" BACKSQL_IDFMT " dn=\"%s\"\n",
+			BACKSQL_IDARG(c_id->eid_id),
+			BACKSQL_IDARG(c_id->eid_keyval),
 			row.cols[ 3 ] );
-#else /* ! BACKSQL_ARBITRARY_KEY */
-		Debug( LDAP_DEBUG_TRACE, "backsql_oc_get_candidates(): "
-			"added entry id=%ld, keyval=%ld dn=\"%s\"\n",
-			c_id->eid_id, c_id->eid_keyval, row.cols[ 3 ] );
-#endif /* ! BACKSQL_ARBITRARY_KEY */
 
 		/* count candidates, for unchecked limit */
 		bsi->bsi_n_candidates--;
@@ -1970,7 +1985,7 @@ cleanup:;
 	SQLFreeStmt( sth, SQL_DROP );
 
 	Debug( LDAP_DEBUG_TRACE, "<==backsql_oc_get_candidates(): %d\n",
-			n_candidates - bsi->bsi_n_candidates, 0, 0 );
+			n_candidates - bsi->bsi_n_candidates );
 
 	return ( bsi->bsi_n_candidates == -1 ? BACKSQL_AVL_STOP : BACKSQL_AVL_CONTINUE );
 }
@@ -1988,7 +2003,9 @@ backsql_search( Operation *op, SlapReply *rs )
 	backsql_srch_info	bsi = { 0 };
 	backsql_entryID		*eid = NULL;
 	struct berval		nbase = BER_BVNULL;
-	unsigned long 		lastid = 0;
+#ifndef BACKSQL_ARBITRARY_KEY
+	ID			lastid = 0;
+#endif /* ! BACKSQL_ARBITRARY_KEY */
 
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_search(): "
 		"base=\"%s\", filter=\"%s\", scope=%d,", 
@@ -2004,7 +2021,7 @@ backsql_search( Operation *op, SlapReply *rs )
 	if ( op->o_req_ndn.bv_len > BACKSQL_MAX_DN_LEN ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_search(): "
 			"search base length (%ld) exceeds max length (%d)\n", 
-			op->o_req_ndn.bv_len, BACKSQL_MAX_DN_LEN, 0 );
+			op->o_req_ndn.bv_len, BACKSQL_MAX_DN_LEN );
 		/*
 		 * FIXME: a LDAP_NO_SUCH_OBJECT could be appropriate
 		 * since it is impossible that such a long DN exists
@@ -2018,8 +2035,7 @@ backsql_search( Operation *op, SlapReply *rs )
 	sres = backsql_get_db_conn( op, &dbh );
 	if ( sres != LDAP_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "backsql_search(): "
-			"could not get connection handle - exiting\n", 
-			0, 0, 0 );
+			"could not get connection handle - exiting\n" );
 		rs->sr_err = sres;
 		rs->sr_text = sres == LDAP_OTHER ?  "SQL-backend error" : NULL;
 		send_ldap_result( op, rs );
@@ -2125,6 +2141,7 @@ backsql_search( Operation *op, SlapReply *rs )
 		( op->ors_limit->lms_s_unchecked == -1 ? -2 :
 		( op->ors_limit->lms_s_unchecked ) ) );
 
+#ifndef BACKSQL_ARBITRARY_KEY
 	/* If paged results are in effect, check the paging cookie */
 	if ( get_pagedresults( op ) > SLAP_CONTROL_IGNORED ) {
 		rs->sr_err = parse_paged_cookie( op, rs );
@@ -2133,6 +2150,7 @@ backsql_search( Operation *op, SlapReply *rs )
 			goto done;
 		}
 	}
+#endif /* ! BACKSQL_ARBITRARY_KEY */
 
 	switch ( bsi.bsi_scope ) {
 	case LDAP_SCOPE_BASE:
@@ -2216,16 +2234,11 @@ backsql_search( Operation *op, SlapReply *rs )
 			goto send_results;
 		}
 
-#ifdef BACKSQL_ARBITRARY_KEY
 		Debug(LDAP_DEBUG_TRACE, "backsql_search(): loading data "
-			"for entry id=%s, oc_id=%ld, keyval=%s\n",
-			eid->eid_id.bv_val, eid->eid_oc_id,
-			eid->eid_keyval.bv_val );
-#else /* ! BACKSQL_ARBITRARY_KEY */
-		Debug(LDAP_DEBUG_TRACE, "backsql_search(): loading data "
-			"for entry id=%ld, oc_id=%ld, keyval=%ld\n",
-			eid->eid_id, eid->eid_oc_id, eid->eid_keyval );
-#endif /* ! BACKSQL_ARBITRARY_KEY */
+			"for entry id=" BACKSQL_IDFMT " oc_id=" BACKSQL_IDNUMFMT ", keyval=" BACKSQL_IDFMT "\n",
+			BACKSQL_IDARG(eid->eid_id),
+			eid->eid_oc_id,
+			BACKSQL_IDARG(eid->eid_keyval) );
 
 		/* check scope */
 		switch ( op->ors_scope ) {
@@ -2275,7 +2288,7 @@ backsql_search( Operation *op, SlapReply *rs )
 			if ( rc != LDAP_SUCCESS ) {
 				Debug( LDAP_DEBUG_TRACE, "backsql_search(): "
 					"error %d in backsql_id2entry() "
-					"- skipping\n", rc, 0, 0 );
+					"- skipping\n", rc );
 				continue;
 			}
 			e = &user_entry;
@@ -2373,7 +2386,7 @@ backsql_search( Operation *op, SlapReply *rs )
 				Debug(LDAP_DEBUG_TRACE, 
 					"backsql_search(): "
 					"has_children failed( %d)\n", 
-					rc, 0, 0 );
+					rc );
 				rc = 1;
 				goto next_entry;
 			}
@@ -2409,6 +2422,7 @@ backsql_search( Operation *op, SlapReply *rs )
 
 		if ( test_filter( op, e, op->ors_filter ) == LDAP_COMPARE_TRUE )
 		{
+#ifndef BACKSQL_ARBITRARY_KEY
 			/* If paged results are in effect, see if the page limit was exceeded */
 			if ( get_pagedresults(op) > SLAP_CONTROL_IGNORED ) {
 				if ( rs->sr_nentries >= ((PagedResultsState *)op->o_pagedresults_state)->ps_size )
@@ -2419,6 +2433,7 @@ backsql_search( Operation *op, SlapReply *rs )
 				}
 				lastid = SQL_TO_PAGECOOKIE( eid->eid_id, eid->eid_oc_id );
 			}
+#endif /* ! BACKSQL_ARBITRARY_KEY */
 			rs->sr_attrs = op->ors_attrs;
 			rs->sr_operational_attrs = NULL;
 			rs->sr_entry = e;
@@ -2438,10 +2453,11 @@ backsql_search( Operation *op, SlapReply *rs )
 				 * better stop
 				 */
 				Debug( LDAP_DEBUG_TRACE, "backsql_search(): "
-					"connection lost\n", 0, 0, 0 );
+					"connection lost\n" );
 				goto end_of_search;
 
 			case LDAP_SIZELIMIT_EXCEEDED:
+			case LDAP_BUSY:
 				goto send_results;
 			}
 		}
@@ -2466,9 +2482,12 @@ end_of_search:;
 
 send_results:;
 	if ( rs->sr_err != SLAPD_ABANDON ) {
+#ifndef BACKSQL_ARBITRARY_KEY
 		if ( get_pagedresults(op) > SLAP_CONTROL_IGNORED ) {
 			send_paged_response( op, rs, NULL );
-		} else {
+		} else
+#endif /* ! BACKSQL_ARBITRARY_KEY */
+		{
 			send_ldap_result( op, rs );
 		}
 	}
@@ -2492,7 +2511,7 @@ send_results:;
 #ifdef BACKSQL_SYNCPROV
 	if ( op->o_sync ) {
 		Operation	op2 = *op;
-		SlapReply	rs2 = { 0 };
+		SlapReply	rs2 = { REP_RESULT };
 		Entry		*e = entry_alloc();
 		slap_callback	cb = { 0 };
 
@@ -2532,7 +2551,7 @@ done:;
 		op->ors_scope = LDAP_SCOPE_BASE;
 	}
 
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_search()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_search()\n" );
 
 	return rs->sr_err;
 }
@@ -2585,8 +2604,7 @@ backsql_entry_get(
 		/* find attribute values */
 		if ( is_entry_alias( bsi.bsi_e ) ) {
 			Debug( LDAP_DEBUG_ACL,
-				"<= backsql_entry_get: entry is an alias\n",
-				0, 0, 0 );
+				"<= backsql_entry_get: entry is an alias\n" );
 			rc = LDAP_ALIAS_PROBLEM;
 			goto return_results;
 		}
@@ -2594,8 +2612,7 @@ backsql_entry_get(
 
 		if ( is_entry_referral( bsi.bsi_e ) ) {
 			Debug( LDAP_DEBUG_ACL,
-				"<= backsql_entry_get: entry is a referral\n",
-				0, 0, 0 );
+				"<= backsql_entry_get: entry is a referral\n" );
 			rc = LDAP_REFERRAL;
 			goto return_results;
 		}
@@ -2603,8 +2620,7 @@ backsql_entry_get(
 		if ( oc && !is_entry_objectclass( bsi.bsi_e, oc, 0 ) ) {
 			Debug( LDAP_DEBUG_ACL,
 					"<= backsql_entry_get: "
-					"failed to find objectClass\n",
-					0, 0, 0 ); 
+					"failed to find objectClass\n" );
 			rc = LDAP_NO_SUCH_ATTRIBUTE;
 			goto return_results;
 		}
@@ -2663,17 +2679,12 @@ backsql_entry_release(
 	return 0;
 }
 
-
+#ifndef BACKSQL_ARBITRARY_KEY
 /* This function is copied verbatim from back-bdb/search.c */
 static int
 parse_paged_cookie( Operation *op, SlapReply *rs )
 {
-	LDAPControl	**c;
 	int		rc = LDAP_SUCCESS;
-	ber_tag_t	tag;
-	ber_int_t	size;
-	BerElement	*ber;
-	struct berval	cookie = BER_BVNULL;
 	PagedResultsState *ps = op->o_pagedresults_state;
 
 	/* this function must be invoked only if the pagedResults
@@ -2681,53 +2692,17 @@ parse_paged_cookie( Operation *op, SlapReply *rs )
 	 * by the frontend */
 	assert( get_pagedresults( op ) > SLAP_CONTROL_IGNORED );
 
-	/* look for the appropriate ctrl structure */
-	for ( c = op->o_ctrls; c[0] != NULL; c++ ) {
-		if ( strcmp( c[0]->ldctl_oid, LDAP_CONTROL_PAGEDRESULTS ) == 0 )
-		{
-			break;
-		}
-	}
-
-	if ( c[0] == NULL ) {
-		rs->sr_text = "missing pagedResults control";
-		return LDAP_PROTOCOL_ERROR;
-	}
-
-	/* Tested by frontend */
-	assert( c[0]->ldctl_value.bv_len > 0 );
-
-	/* Parse the control value
-	 *	realSearchControlValue ::= SEQUENCE {
-	 *		size	INTEGER (0..maxInt),
-	 *				-- requested page size from client
-	 *				-- result set size estimate from server
-	 *		cookie	OCTET STRING
-	 * }
-	 */
-	ber = ber_init( &c[0]->ldctl_value );
-	if ( ber == NULL ) {
-		rs->sr_text = "internal error";
-		return LDAP_OTHER;
-	}
-
-	tag = ber_scanf( ber, "{im}", &size, &cookie );
-
-	/* Tested by frontend */
-	assert( tag != LBER_ERROR );
-	assert( size >= 0 );
-
 	/* cookie decoding/checks deferred to backend... */
-	if ( cookie.bv_len ) {
+	if ( ps->ps_cookieval.bv_len ) {
 		PagedResultsCookie reqcookie;
-		if( cookie.bv_len != sizeof( reqcookie ) ) {
+		if( ps->ps_cookieval.bv_len != sizeof( reqcookie ) ) {
 			/* bad cookie */
 			rs->sr_text = "paged results cookie is invalid";
 			rc = LDAP_PROTOCOL_ERROR;
 			goto done;
 		}
 
-		AC_MEMCPY( &reqcookie, cookie.bv_val, sizeof( reqcookie ));
+		AC_MEMCPY( &reqcookie, ps->ps_cookieval.bv_val, sizeof( reqcookie ));
 
 		if ( reqcookie > ps->ps_cookie ) {
 			/* bad cookie */
@@ -2743,22 +2718,11 @@ parse_paged_cookie( Operation *op, SlapReply *rs )
 
 	} else {
 		/* Initial request.  Initialize state. */
-#if 0
-		if ( op->o_conn->c_pagedresults_state.ps_cookie != 0 ) {
-			/* There's another pagedResults control on the
-			 * same connection; reject new pagedResults controls 
-			 * (allowed by RFC2696) */
-			rs->sr_text = "paged results cookie unavailable; try later";
-			rc = LDAP_UNWILLING_TO_PERFORM;
-			goto done;
-		}
-#endif
 		ps->ps_cookie = 0;
 		ps->ps_count = 0;
 	}
 
 done:;
-	(void)ber_free( ber, 1 );
 
 	return rc;
 }
@@ -2768,7 +2732,7 @@ static void
 send_paged_response( 
 	Operation	*op,
 	SlapReply	*rs,
-	unsigned long	*lastid )
+	ID		*lastid )
 {
 	LDAPControl	ctrl, *ctrls[2];
 	BerElementBuffer berbuf;
@@ -2778,7 +2742,7 @@ send_paged_response(
 
 	Debug(LDAP_DEBUG_ARGS,
 		"send_paged_response: lastid=0x%08lx nentries=%d\n", 
-		lastid ? *lastid : 0, rs->sr_nentries, NULL );
+		lastid ? *lastid : 0, rs->sr_nentries );
 
 	BER_BVZERO( &ctrl.ldctl_value );
 	ctrls[0] = &ctrl;
@@ -2819,3 +2783,4 @@ send_paged_response(
 done:
 	(void) ber_free_buf( ber );
 }
+#endif /* ! BACKSQL_ARBITRARY_KEY */
