@@ -1,8 +1,8 @@
 /* mr.c - routines to manage matching rule definitions */
-/* $OpenLDAP: pkg/ldap/servers/slapd/mr.c,v 1.54.2.2 2003/12/01 00:58:35 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2003 The OpenLDAP Foundation.
+ * Copyright 1998-2021 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -23,7 +23,6 @@
 #include <ac/socket.h>
 
 #include "slap.h"
-#include "ldap_pvt.h"
 
 struct mindexrec {
 	struct berval	mir_name;
@@ -31,9 +30,9 @@ struct mindexrec {
 };
 
 static Avlnode	*mr_index = NULL;
-static LDAP_SLIST_HEAD(MRList, slap_matching_rule) mr_list
+static LDAP_SLIST_HEAD(MRList, MatchingRule) mr_list
 	= LDAP_SLIST_HEAD_INITIALIZER(&mr_list);
-static LDAP_SLIST_HEAD(MRUList, slap_matching_rule_use) mru_list
+static LDAP_SLIST_HEAD(MRUList, MatchingRuleUse) mru_list
 	= LDAP_SLIST_HEAD_INITIALIZER(&mru_list);
 
 static int
@@ -46,7 +45,7 @@ mr_index_cmp(
 	const struct mindexrec	*mir2 = v_mir2;
 	int i = mir1->mir_name.bv_len - mir2->mir_name.bv_len;
 	if (i) return i;
-	return (strcmp( mir1->mir_name.bv_val, mir2->mir_name.bv_val ));
+	return (strcasecmp( mir1->mir_name.bv_val, mir2->mir_name.bv_val ));
 }
 
 static int
@@ -59,7 +58,7 @@ mr_index_name_cmp(
 	const struct mindexrec *mir  = v_mir;
 	int i = name->bv_len - mir->mir_name.bv_len;
 	if (i) return i;
-	return (strncmp( name->bv_val, mir->mir_name.bv_val, name->bv_len ));
+	return (strncasecmp( name->bv_val, mir->mir_name.bv_val, name->bv_len ));
 }
 
 MatchingRule *
@@ -146,6 +145,67 @@ mr_insert(
 	return 0;
 }
 
+int
+mr_make_syntax_compat_with_mr(
+	Syntax		*syn,
+	MatchingRule	*mr )
+{
+	int		n = 0;
+
+	assert( syn != NULL );
+	assert( mr != NULL );
+
+	if ( mr->smr_compat_syntaxes ) {
+		/* count existing */
+		for ( n = 0;
+			mr->smr_compat_syntaxes[ n ];
+			n++ )
+		{
+			if ( mr->smr_compat_syntaxes[ n ] == syn ) {
+				/* already compatible; mmmmh... */
+				return 1;
+			}
+		}
+	}
+
+	mr->smr_compat_syntaxes = ch_realloc(
+		mr->smr_compat_syntaxes,
+		sizeof( Syntax * )*(n + 2) );
+	mr->smr_compat_syntaxes[ n ] = syn;
+	mr->smr_compat_syntaxes[ n + 1 ] = NULL;
+
+	return 0;
+}
+
+int
+mr_make_syntax_compat_with_mrs(
+	const char *syntax,
+	char *const *mrs )
+{
+	int	r, rc = 0;
+	Syntax	*syn;
+
+	assert( syntax != NULL );
+	assert( mrs != NULL );
+
+	syn = syn_find( syntax );
+	if ( syn == NULL ) {
+		return -1;
+	}
+
+	for ( r = 0; mrs[ r ] != NULL; r++ ) {
+		MatchingRule	*mr = mr_find( mrs[ r ] );
+		if ( mr == NULL ) {
+			/* matchingRule not found -- ignore by now */
+			continue;
+		}
+
+		rc += mr_make_syntax_compat_with_mr( syn, mr );
+	}
+
+	return rc;
+}
+
 int
 mr_add(
     LDAPMatchingRule		*mr,
@@ -170,6 +230,7 @@ mr_add(
 		for( i=0; def->mrd_compat_syntaxes[i]; i++ ) {
 			compat_syn[i] = syn_find( def->mrd_compat_syntaxes[i] );
 			if( compat_syn[i] == NULL ) {
+				ch_free( compat_syn );
 				return SLAP_SCHERR_SYN_NOT_FOUND;
 			}
 		}
@@ -199,10 +260,12 @@ mr_add(
 			smr->smr_syntax = syn;
 		} else {
 			*err = smr->smr_syntax_oid;
+			ch_free( smr );
 			return SLAP_SCHERR_SYN_NOT_FOUND;
 		}
 	} else {
 		*err = "";
+		ch_free( smr );
 		return SLAP_SCHERR_MR_INCOMPLETE;
 	}
 	code = mr_insert(smr,err);
@@ -218,34 +281,19 @@ register_matching_rule(
 	int		code;
 	const char	*err;
 
-	if( def->mrd_usage == SLAP_MR_NONE &&
-		def->mrd_compat_syntaxes == NULL )
-	{
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, ERR, 
-			"register_matching_rule: %s not usable\n", def->mrd_desc, 0, 0 );
-#else
+	if( def->mrd_usage == SLAP_MR_NONE && def->mrd_compat_syntaxes == NULL ) {
 		Debug( LDAP_DEBUG_ANY, "register_matching_rule: not usable %s\n",
-		    def->mrd_desc, 0, 0 );
-#endif
+		    def->mrd_desc );
 
 		return -1;
 	}
 
 	if( def->mrd_associated != NULL ) {
 		amr = mr_find( def->mrd_associated );
-
 		if( amr == NULL ) {
-#ifdef NEW_LOGGING
-			LDAP_LOG( OPERATION, ERR,
-			   "register_matching_rule: could not locate associated "
-			   "matching rule %s for %s\n",
-				def->mrd_associated, def->mrd_desc, 0 );
-#else
-			Debug( LDAP_DEBUG_ANY, "register_matching_rule: could not locate "
-				"associated matching rule %s for %s\n",
-				def->mrd_associated, def->mrd_desc, 0 );
-#endif
+			Debug( LDAP_DEBUG_ANY, "register_matching_rule: "
+				"could not locate associated matching rule %s for %s\n",
+				def->mrd_associated, def->mrd_desc );
 
 			return -1;
 		}
@@ -256,32 +304,16 @@ register_matching_rule(
 			if (( def->mrd_usage & SLAP_MR_EQUALITY ) &&
 				(( def->mrd_usage & SLAP_MR_SUBTYPE_MASK ) != SLAP_MR_NONE ))
 			{
-#ifdef NEW_LOGGING
-				LDAP_LOG( OPERATION, ERR,
-				   "register_matching_rule: inappropriate (approx) association "
-						"%s for %s\n",
-					def->mrd_associated, def->mrd_desc, 0 );
-#else
-				Debug( LDAP_DEBUG_ANY,
-				   "register_matching_rule: inappropriate (approx) association "
-						"%s for %s\n",
-					def->mrd_associated, def->mrd_desc, 0 );
-#endif
+				Debug( LDAP_DEBUG_ANY, "register_matching_rule: "
+						"inappropriate (approx) association %s for %s\n",
+					def->mrd_associated, def->mrd_desc );
 				return -1;
 			}
 
 		} else if (!( amr->smr_usage & SLAP_MR_EQUALITY )) {
-#ifdef NEW_LOGGING
-				LDAP_LOG( OPERATION, ERR,
-				   "register_matching_rule: inappropriate (equalilty) association "
-						"%s for %s\n",
-					def->mrd_associated, def->mrd_desc, 0 );
-#else
-				Debug( LDAP_DEBUG_ANY,
-				   "register_matching_rule: inappropriate (equalilty) association "
-						"%s for %s\n",
-					def->mrd_associated, def->mrd_desc, 0 );
-#endif
+				Debug( LDAP_DEBUG_ANY, "register_matching_rule: "
+					"inappropriate (equalilty) association %s for %s\n",
+					def->mrd_associated, def->mrd_desc );
 				return -1;
 		}
 	}
@@ -289,15 +321,9 @@ register_matching_rule(
 	mr = ldap_str2matchingrule( def->mrd_desc, &code, &err,
 		LDAP_SCHEMA_ALLOW_ALL );
 	if ( !mr ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, ERR, 
-			"register_matching_rule: %s before %s in %s.\n",
-			ldap_scherr2str(code), err, def->mrd_desc );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"Error in register_matching_rule: %s before %s in %s\n",
 		    ldap_scherr2str(code), err, def->mrd_desc );
-#endif
 
 		return -1;
 	}
@@ -308,15 +334,9 @@ register_matching_rule(
 	ldap_memfree( mr );
 
 	if ( code ) {
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, ERR, 
-			"register_matching_rule: %s for %s in %s.\n",
-			scherr2str(code), err, def->mrd_desc );
-#else
 		Debug( LDAP_DEBUG_ANY,
 			"Error in register_matching_rule: %s for %s in %s\n",
 		    scherr2str(code), err, def->mrd_desc );
-#endif
 
 		return -1;
 	}
@@ -353,15 +373,12 @@ matching_rule_use_init( void )
 	MatchingRule	*mr;
 	MatchingRuleUse	**mru_ptr = &LDAP_SLIST_FIRST(&mru_list);
 
-#ifdef NEW_LOGGING
-	LDAP_LOG( OPERATION, INFO, "matching_rule_use_init\n", 0, 0, 0 );
-#else
-	Debug( LDAP_DEBUG_TRACE, "matching_rule_use_init\n", 0, 0, 0 );
-#endif
+	Debug( LDAP_DEBUG_TRACE, "matching_rule_use_init\n" );
 
 	LDAP_SLIST_FOREACH( mr, &mr_list, smr_next ) {
 		AttributeType	*at;
-		MatchingRuleUse	mru_storage, *mru = &mru_storage;
+		MatchingRuleUse	mru_storage = {{ 0 }},
+				*mru = &mru_storage;
 
 		char		**applies_oids = NULL;
 
@@ -386,8 +403,6 @@ matching_rule_use_init( void )
 			continue;
 		}
 
-		memset( mru, 0, sizeof( MatchingRuleUse ) );
-
 		/*
 		 * Note: we're using the same values of the corresponding 
 		 * MatchingRule structure; maybe we'd copy them ...
@@ -400,15 +415,9 @@ matching_rule_use_init( void )
 		mru->smru_names = mr->smr_names;
 		mru->smru_desc = mr->smr_desc;
 
-#ifdef NEW_LOGGING
-		LDAP_LOG( OPERATION, INFO, "    %s (%s): ", 
-				mru->smru_oid, 
-				mru->smru_names ? mru->smru_names[ 0 ] : "", 0 );
-#else
 		Debug( LDAP_DEBUG_TRACE, "    %s (%s): ", 
 				mru->smru_oid, 
-				mru->smru_names ? mru->smru_names[ 0 ] : "", 0 );
-#endif
+				mru->smru_names ? mru->smru_names[ 0 ] : "" );
 
 		at = NULL;
 		for ( at_start( &at ); at; at_next( &at ) ) {
@@ -426,19 +435,11 @@ matching_rule_use_init( void )
 		 */
 		if ( applies_oids != NULL ) {
 			mru->smru_applies_oids = applies_oids;
-#ifdef NEW_LOGGING
 			{
 				char *str = ldap_matchingruleuse2str( &mru->smru_mruleuse );
-				LDAP_LOG( OPERATION, INFO, "matchingRuleUse: %s\n", str, 0, 0 );
+				Debug( LDAP_DEBUG_TRACE, "matchingRuleUse: %s\n", str );
 				ldap_memfree( str );
 			}
-#else
-			{
-				char *str = ldap_matchingruleuse2str( &mru->smru_mruleuse );
-				Debug( LDAP_DEBUG_TRACE, "matchingRuleUse: %s\n", str, 0, 0 );
-				ldap_memfree( str );
-			}
-#endif
 
 			mru = (MatchingRuleUse *)ber_memalloc( sizeof( MatchingRuleUse ) );
 			/* call-forward from MatchingRule to MatchingRuleUse */
@@ -455,13 +456,16 @@ matching_rule_use_init( void )
 	return( 0 );
 }
 
-int mr_usable_with_at(
-	MatchingRule *mr,
-	AttributeType *at )
+int
+mr_usable_with_at(
+	MatchingRule	*mr,
+	AttributeType	*at )
 {
-	if( mr->smr_usage & SLAP_MR_EXT && ( 
+	if ( ( mr->smr_usage & SLAP_MR_EXT ) && (
 		mr->smr_syntax == at->sat_syntax ||
-		mr == at->sat_equality || mr == at->sat_approx ) )
+		mr == at->sat_equality ||
+		mr == at->sat_approx ||
+		syn_is_sup( at->sat_syntax, mr->smr_syntax ) ) )
 	{
 		return 1;
 	}
@@ -501,7 +505,7 @@ int mr_schema_info( Entry *e )
 		}
 #if 0
 		Debug( LDAP_DEBUG_TRACE, "Merging mr [%lu] %s\n",
-			mr->smr_str.bv_len, mr->smr_str.bv_val, 0 );
+			mr->smr_str.bv_len, mr->smr_str.bv_val );
 #endif
 
 		nval.bv_val = mr->smr_oid;
@@ -532,7 +536,7 @@ int mru_schema_info( Entry *e )
 
 #if 0
 		Debug( LDAP_DEBUG_TRACE, "Merging mru [%lu] %s\n",
-			mru->smru_str.bv_len, mru->smru_str.bv_val, 0 );
+			mru->smru_str.bv_len, mru->smru_str.bv_val );
 #endif
 
 		nval.bv_val = mru->smru_oid;
