@@ -1,8 +1,8 @@
 /* unique.c - attribute uniqueness module */
-/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/unique.c,v 1.20.2.3 2007/08/31 23:14:06 quanah Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2004-2007 The OpenLDAP Foundation.
+ * Copyright 2004-2021 The OpenLDAP Foundation.
  * Portions Copyright 2004,2006-2007 Symas Corporation.
  * All rights reserved.
  *
@@ -17,7 +17,7 @@
 /* ACKNOWLEDGEMENTS: 
  * This work was initially developed by Symas Corporation for
  * inclusion in OpenLDAP Software, with subsequent enhancements by
- * Matthew Backes at Symas Corporation.  This work was sponsored by
+ * Emily Backes at Symas Corporation.  This work was sponsored by
  * Hewlett-Packard.
  */
 
@@ -31,7 +31,7 @@
 #include <ac/socket.h>
 
 #include "slap.h"
-#include "config.h"
+#include "slap-config.h"
 
 #define UNIQUE_DEFAULT_URI ("ldap:///??sub")
 
@@ -44,25 +44,28 @@ typedef struct unique_attrs_s {
 
 typedef struct unique_domain_uri_s {
 	struct unique_domain_uri_s *next;
-	struct berval *dn;
-	struct berval *ndn;
-	struct berval *filter;
+	struct berval dn;
+	struct berval ndn;
+	struct berval filter;
+	Filter *f;
 	struct unique_attrs_s *attrs;
 	int scope;
 } unique_domain_uri;
 
 typedef struct unique_domain_s {
 	struct unique_domain_s *next;
-	struct berval *domain_spec;
+	struct berval domain_spec;
 	struct unique_domain_uri_s *uri;
 	char ignore;                          /* polarity of attributes */
 	char strict;                          /* null considered unique too */
+	char serial;						/* serialize execution */
 } unique_domain;
 
 typedef struct unique_data_s {
 	struct unique_domain_s *domains;
 	struct unique_domain_s *legacy;
 	char legacy_strict_set;
+	ldap_pvt_thread_mutex_t	serial_mutex;
 } unique_data;
 
 typedef struct unique_counter_s {
@@ -75,7 +78,7 @@ enum {
 	UNIQUE_IGNORE,
 	UNIQUE_ATTR,
 	UNIQUE_STRICT,
-	UNIQUE_URI
+	UNIQUE_URI,
 };
 
 static ConfigDriver unique_cf_base;
@@ -138,9 +141,10 @@ unique_free_domain_uri ( unique_domain_uri *uri )
 
 	while ( uri ) {
 		next_uri = uri->next;
-		ber_bvfree ( uri->dn );
-		ber_bvfree ( uri->ndn );
-		ber_bvfree ( uri->filter );
+		ch_free ( uri->dn.bv_val );
+		ch_free ( uri->ndn.bv_val );
+		ch_free ( uri->filter.bv_val );
+		filter_free( uri->f );
 		attr = uri->attrs;
 		while ( attr ) {
 			next_attr = attr->next;
@@ -160,7 +164,7 @@ unique_free_domain ( unique_domain *domain )
 
 	while ( domain ) {
 		next_domain = domain->next;
-		ber_bvfree ( domain->domain_spec );
+		ch_free ( domain->domain_spec.bv_val );
 		unique_free_domain_uri ( domain->uri );
 		ch_free ( domain );
 		domain = next_domain;
@@ -182,12 +186,20 @@ unique_new_domain_uri ( unique_domain_uri **urip,
 
 	uri = ch_calloc ( 1, sizeof ( unique_domain_uri ) );
 
+	if ( url_desc->lud_host && url_desc->lud_host[0] ) {
+		snprintf( c->cr_msg, sizeof( c->cr_msg ),
+			  "host <%s> not allowed in URI",
+			  url_desc->lud_host );
+		rc = ARG_BAD_CONF;
+		goto exit;
+	}
+
 	if ( url_desc->lud_dn && url_desc->lud_dn[0] ) {
-		ber_str2bv( url_desc->lud_dn, 0, 1, &bv );
+		ber_str2bv( url_desc->lud_dn, 0, 0, &bv );
 		rc = dnPrettyNormal( NULL,
 				     &bv,
-				     uri->dn,
-				     uri->ndn,
+				     &uri->dn,
+				     &uri->ndn,
 				     NULL );
 		if ( rc != LDAP_SUCCESS ) {
 			snprintf( c->cr_msg, sizeof( c->cr_msg ),
@@ -197,14 +209,30 @@ unique_new_domain_uri ( unique_domain_uri **urip,
 			goto exit;
 		}
 
-		if ( !dnIsSuffix ( uri->ndn, &be->be_nsuffix[0] ) ) {
-			sprintf ( c->cr_msg,
+		if ( be->be_nsuffix == NULL ) {
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				  "suffix must be set" );
+			Debug ( LDAP_DEBUG_CONFIG, "unique config: %s\n",
+				c->cr_msg );
+			rc = ARG_BAD_CONF;
+			goto exit;
+		}
+
+		if ( !dnIsSuffix ( &uri->ndn, &be->be_nsuffix[0] ) ) {
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
 				  "dn <%s> is not a suffix of backend base dn <%s>",
-				  uri->dn->bv_val,
+				  uri->dn.bv_val,
 				  be->be_nsuffix[0].bv_val );
 			rc = ARG_BAD_CONF;
 			goto exit;
 		}
+
+		if ( BER_BVISNULL( &be->be_rootndn ) || BER_BVISEMPTY( &be->be_rootndn ) ) {
+			Debug( LDAP_DEBUG_ANY,
+				"slapo-unique needs a rootdn; "
+				"backend <%s> has none, YMMV.\n",
+				be->be_nsuffix[0].bv_val );
+		}
 	}
 
 	attr_str = url_desc->lud_attrs;
@@ -238,24 +266,31 @@ unique_new_domain_uri ( unique_domain_uri **urip,
 	}
 
 	if (url_desc->lud_filter) {
-		Filter * f;
-		uri->filter = ber_str2bv( url_desc->lud_filter, 0, 1, NULL);
-		f = str2filter( uri->filter->bv_val );
-		if ( !f ) {
+		char *ptr;
+		uri->f = str2filter( url_desc->lud_filter );
+		if ( !uri->f ) {
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				  "unique: bad filter");
+			rc = ARG_BAD_CONF;
+			goto exit;
+		}
+		/* make sure the strfilter is in normal form (ITS#5581) */
+		filter2bv( uri->f, &uri->filter );
+		ptr = strstr( uri->filter.bv_val, "(?=" /*)*/ );
+		if ( ptr != NULL && ptr <= ( uri->filter.bv_val - STRLENOF( "(?=" /*)*/ ) + uri->filter.bv_len ) )
+		{
 			snprintf( c->cr_msg, sizeof( c->cr_msg ),
 				  "unique: bad filter");
 			rc = ARG_BAD_CONF;
 			goto exit;
 		}
-		filter_free( f );
 	}
 exit:
-	if ( bv.bv_val ) ber_memfree ( bv.bv_val );
 	uri->next = *urip;
 	*urip = uri;
 	if ( rc ) {
 		Debug ( LDAP_DEBUG_CONFIG|LDAP_DEBUG_NONE,
-			"%s: %s\n", c->log, c->cr_msg, 0 );
+			"%s: %s\n", c->log, c->cr_msg );
 		unique_free_domain_uri ( uri );
 		*urip = NULL;
 	}
@@ -282,7 +317,7 @@ unique_new_domain_uri_basic ( unique_domain_uri **urip,
  *
  * domain_specs look like
  *
- * [strict ][ignore ]uri[[ uri]...]
+ * [strict ][ignore ][serialize ]uri[[ uri]...]
  * e.g. "ldap:///ou=foo,o=bar?uid?sub ldap:///ou=baz,o=bar?uid?sub"
  *      "strict ldap:///ou=accounts,o=bar?uid,uidNumber?one"
  *      etc
@@ -302,10 +337,10 @@ unique_new_domain ( unique_domain **domainp,
 	LDAPURLDesc *url_desc, *url_descs = NULL;
 
 	Debug(LDAP_DEBUG_TRACE, "==> unique_new_domain <%s>\n",
-	      domain_spec, 0, 0);
+	      domain_spec );
 
 	domain = ch_calloc ( 1, sizeof (unique_domain) );
-	domain->domain_spec = ber_str2bv( domain_spec, 0, 1, NULL );
+	ber_str2bv( domain_spec, 0, 1, &domain->domain_spec );
 
 	uri_start = domain_spec;
 	if ( strncasecmp ( uri_start, "ignore ",
@@ -313,6 +348,11 @@ unique_new_domain ( unique_domain **domainp,
 		domain->ignore = 1;
 		uri_start += STRLENOF( "ignore " );
 	}
+	if ( strncasecmp ( uri_start, "serialize ",
+			   STRLENOF( "serialize " ) ) == 0 ) {
+		domain->serial = 1;
+		uri_start += STRLENOF( "serialize " );
+	}
 	if ( strncasecmp ( uri_start, "strict ",
 			   STRLENOF( "strict " ) ) == 0 ) {
 		domain->strict = 1;
@@ -335,7 +375,7 @@ unique_new_domain ( unique_domain **domainp,
 
 	for ( url_desc = url_descs;
 	      url_desc;
-	      url_desc = url_descs->lud_next ) {
+	      url_desc = url_desc->lud_next ) {
 		rc = unique_new_domain_uri ( &domain->uri,
 					     url_desc,
 					     c );
@@ -352,7 +392,7 @@ exit:
 	*domainp = domain;
 	if ( rc ) {
 		Debug ( LDAP_DEBUG_CONFIG|LDAP_DEBUG_NONE,
-			"%s: %s\n", c->log, c->cr_msg, 0 );
+			"%s: %s\n", c->log, c->cr_msg );
 		unique_free_domain ( domain );
 		*domainp = NULL;
 	}
@@ -372,24 +412,23 @@ unique_cf_base( ConfigArgs *c )
 	switch ( c->op ) {
 	case SLAP_CONFIG_EMIT:
 		rc = 0;
-		if ( legacy && legacy->uri && legacy->uri->dn ) {
+		if ( legacy && legacy->uri && legacy->uri->dn.bv_val ) {
 			rc = value_add_one ( &c->rvalue_vals,
-					     legacy->uri->dn );
+					     &legacy->uri->dn );
 			if ( rc ) return rc;
 			rc = value_add_one ( &c->rvalue_nvals,
-					     legacy->uri->ndn );
+					     &legacy->uri->ndn );
 			if ( rc ) return rc;
 		}
 		break;
 	case LDAP_MOD_DELETE:
-		assert ( legacy && legacy->uri && legacy->uri->dn );
+		assert ( legacy && legacy->uri && legacy->uri->dn.bv_val );
 		rc = 0;
-		ber_bvfree ( legacy->uri->dn );
-		ber_bvfree ( legacy->uri->ndn );
-		legacy->uri->dn = NULL;
-		legacy->uri->ndn = NULL;
-		if ( !legacy->uri->attrs
-		     && !legacy->uri->dn ) {
+		ch_free ( legacy->uri->dn.bv_val );
+		ch_free ( legacy->uri->ndn.bv_val );
+		BER_BVZERO( &legacy->uri->dn );
+		BER_BVZERO( &legacy->uri->ndn );
+		if ( !legacy->uri->attrs ) {
 			unique_free_domain_uri ( legacy->uri );
 			legacy->uri = NULL;
 		}
@@ -401,19 +440,27 @@ unique_cf_base( ConfigArgs *c )
 	case LDAP_MOD_ADD:
 	case SLAP_CONFIG_ADD:
 		if ( domains ) {
-			sprintf ( c->cr_msg,
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
 				  "cannot set legacy attrs when URIs are present" );
 			Debug ( LDAP_DEBUG_CONFIG, "unique config: %s\n",
-				c->cr_msg, NULL, NULL );
+				c->cr_msg );
+			rc = ARG_BAD_CONF;
+			break;
+		}
+		if ( be->be_nsuffix == NULL ) {
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				  "suffix must be set" );
+			Debug ( LDAP_DEBUG_CONFIG, "unique config: %s\n",
+				c->cr_msg );
 			rc = ARG_BAD_CONF;
 			break;
 		}
 		if ( !dnIsSuffix ( &c->value_ndn,
 				   &be->be_nsuffix[0] ) ) {
-			sprintf ( c->cr_msg,
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
 				  "dn is not a suffix of backend base" );
 			Debug ( LDAP_DEBUG_CONFIG, "unique config: %s\n",
-				c->cr_msg, NULL, NULL );
+				c->cr_msg );
 			rc = ARG_BAD_CONF;
 			break;
 		}
@@ -425,16 +472,23 @@ unique_cf_base( ConfigArgs *c )
 		}
 		if ( !legacy->uri )
 			unique_new_domain_uri_basic ( &legacy->uri, c );
-		ber_bvfree ( legacy->uri->dn );
-		ber_bvfree ( legacy->uri->ndn );
-		legacy->uri->dn = ber_bvdup ( &c->value_dn );
-		legacy->uri->ndn = ber_bvdup ( &c->value_ndn );
+		ch_free ( legacy->uri->dn.bv_val );
+		ch_free ( legacy->uri->ndn.bv_val );
+		legacy->uri->dn = c->value_dn;
+		legacy->uri->ndn = c->value_ndn;
 		rc = 0;
 		break;
 	default:
 		abort();
 	}
 
+	if ( rc ) {
+		ch_free( c->value_dn.bv_val );
+		BER_BVZERO( &c->value_dn );
+		ch_free( c->value_ndn.bv_val );
+		BER_BVZERO( &c->value_ndn );
+	}
+
 	return rc;
 }
 
@@ -485,7 +539,7 @@ unique_cf_attrs( ConfigArgs *c )
 				ch_free (attr);
 			}
 			if ( !legacy->uri->attrs
-			     && !legacy->uri->dn ) {
+			     && !legacy->uri->dn.bv_val ) {
 				unique_free_domain_uri ( legacy->uri );
 				legacy->uri = NULL;
 			}
@@ -499,10 +553,10 @@ unique_cf_attrs( ConfigArgs *c )
 	case LDAP_MOD_ADD:
 	case SLAP_CONFIG_ADD:
 		if ( domains ) {
-			sprintf ( c->cr_msg,
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
 				  "cannot set legacy attrs when URIs are present" );
 			Debug ( LDAP_DEBUG_CONFIG, "unique config: %s\n",
-				c->cr_msg, NULL, NULL );
+				c->cr_msg );
 			rc = ARG_BAD_CONF;
 			break;
 		}
@@ -510,10 +564,10 @@ unique_cf_attrs( ConfigArgs *c )
 		     && legacy->uri
 		     && legacy->uri->attrs
 		     && (c->type == UNIQUE_IGNORE) != legacy->ignore ) {
-			sprintf ( c->cr_msg,
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
 				  "cannot set both attrs and ignore-attrs" );
 			Debug ( LDAP_DEBUG_CONFIG, "unique config: %s\n",
-				c->cr_msg, NULL, NULL );
+				c->cr_msg );
 			rc = ARG_BAD_CONF;
 			break;
 		}
@@ -575,7 +629,7 @@ unique_cf_attrs( ConfigArgs *c )
 
 	if ( rc ) {
 		Debug ( LDAP_DEBUG_CONFIG|LDAP_DEBUG_NONE,
-			"%s: %s\n", c->log, c->cr_msg, 0 );
+			"%s: %s\n", c->log, c->cr_msg );
 	}
 	return rc;
 }
@@ -597,11 +651,7 @@ unique_cf_strict( ConfigArgs *c )
 		 * and missing is necessary to add olcUniqueURIs...
 		 */
 		if ( private->legacy_strict_set ) {
-			struct berval bv;
-			bv.bv_val = legacy->strict ? "TRUE" : "FALSE";
-			bv.bv_len = legacy->strict ?
-				STRLENOF("TRUE") :
-				STRLENOF("FALSE");
+			struct berval bv = legacy->strict ? slap_true_bv : slap_false_bv;
 			value_add_one ( &c->rvalue_vals, &bv );
 		}
 		rc = 0;
@@ -620,10 +670,10 @@ unique_cf_strict( ConfigArgs *c )
 	case LDAP_MOD_ADD:
 	case SLAP_CONFIG_ADD:
 		if ( domains ) {
-			sprintf ( c->cr_msg,
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
 				  "cannot set legacy attrs when URIs are present" );
 			Debug ( LDAP_DEBUG_CONFIG, "unique config: %s\n",
-				c->cr_msg, NULL, NULL );
+				c->cr_msg );
 			rc = ARG_BAD_CONF;
 			break;
 		}
@@ -663,10 +713,7 @@ unique_cf_uri( ConfigArgs *c )
 		      domain;
 		      domain = domain->next ) {
 			rc = value_add_one ( &c->rvalue_vals,
-					     domain->domain_spec );
-			if ( rc ) break;
-			rc = value_add_one ( &c->rvalue_nvals,
-					     domain->domain_spec );
+					     &domain->domain_spec );
 			if ( rc ) break;
 		}
 		break;
@@ -702,13 +749,13 @@ unique_cf_uri( ConfigArgs *c )
 		rc = 0;
 		break;
 
-	case SLAP_CONFIG_ADD: /* fallthrough */
+	case SLAP_CONFIG_ADD: /* fallthru */
 	case LDAP_MOD_ADD:
 		if ( legacy ) {
-			sprintf ( c->cr_msg,
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
 				  "cannot set Uri when legacy attrs are present" );
 			Debug ( LDAP_DEBUG_CONFIG, "unique config: %s\n",
-				c->cr_msg, NULL, NULL );
+				c->cr_msg );
 			rc = ARG_BAD_CONF;
 			break;
 		}
@@ -745,11 +792,13 @@ unique_db_init(
 )
 {
 	slap_overinst *on = (slap_overinst *)be->bd_info;
-	unique_data **privatep = (unique_data **) &on->on_bi.bi_private;
+	unique_data *private;
 
-	Debug(LDAP_DEBUG_TRACE, "==> unique_db_init\n", 0, 0, 0);
+	Debug(LDAP_DEBUG_TRACE, "==> unique_db_init\n" );
 
-	*privatep = ch_calloc ( 1, sizeof ( unique_data ) );
+	private = ch_calloc ( 1, sizeof ( unique_data ) );
+	ldap_pvt_thread_mutex_init( &private->serial_mutex );
+	on->on_bi.bi_private = private;
 
 	return 0;
 }
@@ -761,10 +810,9 @@ unique_db_destroy(
 )
 {
 	slap_overinst *on = (slap_overinst *)be->bd_info;
-	unique_data **privatep = (unique_data **) &on->on_bi.bi_private;
-	unique_data *private = *privatep;
+	unique_data *private = on->on_bi.bi_private;
 
-	Debug(LDAP_DEBUG_TRACE, "==> unique_db_destroy\n", 0, 0, 0);
+	Debug(LDAP_DEBUG_TRACE, "==> unique_db_destroy\n" );
 
 	if ( private ) {
 		unique_domain *domains = private->domains;
@@ -772,54 +820,14 @@ unique_db_destroy(
 
 		unique_free_domain ( domains );
 		unique_free_domain ( legacy );
+		ldap_pvt_thread_mutex_destroy( &private->serial_mutex );
 		ch_free ( private );
-		*privatep = NULL;
+		on->on_bi.bi_private = NULL;
 	}
 
 	return 0;
 }
 
-static int
-unique_open(
-	BackendDB *be,
-	ConfigReply *cr
-)
-{
-	Debug(LDAP_DEBUG_TRACE, "unique_open: overlay initialized\n", 0, 0, 0);
-
-	return 0;
-}
-
-
-/*
-** Leave unique_data but wipe out config
-**
-*/
-
-static int
-unique_close(
-	BackendDB *be,
-	ConfigReply *cr
-)
-{
-	slap_overinst *on	= (slap_overinst *) be->bd_info;
-	unique_data **privatep = (unique_data **) &on->on_bi.bi_private;
-	unique_data *private = *privatep;
-
-	Debug(LDAP_DEBUG_TRACE, "==> unique_close\n", 0, 0, 0);
-
-	if ( private ) {
-		unique_domain *domains = private->domains;
-		unique_domain *legacy = private->legacy;
-
-		unique_free_domain ( domains );
-		unique_free_domain ( legacy );
-		memset ( private, 0, sizeof ( unique_data ) );
-	}
-
-	return ( 0 );
-}
-
 
 /*
 ** search callback
@@ -846,7 +854,7 @@ static int count_attr_cb(
 	if ( dn_match( uc->ndn, &rs->sr_entry->e_nname )) return(0);
 
 	Debug(LDAP_DEBUG_TRACE, "==> count_attr_cb <%s>\n",
-		rs->sr_entry ? rs->sr_entry->e_name.bv_val : "UNKNOWN_DN", 0, 0);
+		rs->sr_entry ? rs->sr_entry->e_name.bv_val : "UNKNOWN_DN" );
 
 	uc->count++;
 
@@ -902,6 +910,7 @@ build_filter(
 	AttributeDescription *ad,
 	BerVarray b,
 	char *kp,
+	int ks,
 	void *ctx
 )
 {
@@ -923,15 +932,23 @@ build_filter(
 		if ( b && b[0].bv_val ) {
 			for ( i = 0; b[i].bv_val; i++ ) {
 				struct berval	bv;
+				int len;
 
 				ldap_bv2escaped_filter_value_x( &b[i], &bv, 1, ctx );
-				kp += sprintf( kp, "(%s=%s)", ad->ad_cname.bv_val, bv.bv_val );
+				if (!b[i].bv_len)
+					bv.bv_val = b[i].bv_val;
+				len = snprintf( kp, ks, "(%s=%s)", ad->ad_cname.bv_val, bv.bv_val );
+				assert( len >= 0 && len < ks );
+				kp += len;
 				if ( bv.bv_val != b[i].bv_val ) {
 					ber_memfree_x( bv.bv_val, ctx );
 				}
 			}
 		} else if ( domain->strict ) {
-			kp += sprintf( kp, "(%s=*)", ad->ad_cname.bv_val );
+			int len;
+			len = snprintf( kp, ks, "(%s=*)", ad->ad_cname.bv_val );
+			assert( len >= 0 && len < ks );
+			kp += len;
 		}
 		break;
 	}
@@ -945,7 +962,7 @@ unique_search(
 	struct berval * dn,
 	int scope,
 	SlapReply *rs,
-	char *key
+	struct berval *key
 )
 {
 	slap_overinst *on = (slap_overinst *) op->o_bd->bd_info;
@@ -953,11 +970,20 @@ unique_search(
 	slap_callback cb = { NULL, NULL, NULL, NULL }; /* XXX */
 	unique_counter uq = { NULL, 0 };
 	int rc;
+	char *errmsg;
+	int errmsgsize;
 
-	Debug(LDAP_DEBUG_TRACE, "==> unique_search %s\n", key, 0, 0);
+	Debug(LDAP_DEBUG_TRACE, "==> unique_search %s\n", key->bv_val );
 
-	nop->ors_filter = str2filter_x(nop, key);
-	ber_str2bv(key, 0, 0, &nop->ors_filterstr);
+	nop->ors_filter = str2filter_x(nop, key->bv_val);
+	if(nop->ors_filter == NULL) {
+		op->o_bd->bd_info = (BackendInfo *) on->on_info;
+		send_ldap_error(op, rs, LDAP_OTHER,
+			"unique_search invalid filter");
+		return(rs->sr_err);
+	}
+
+	nop->ors_filterstr = *key;
 
 	cb.sc_response	= (slap_response*)count_attr_cb;
 	cb.sc_private	= &uq;
@@ -978,25 +1004,45 @@ unique_search(
 
 	nop->o_bd = on->on_info->oi_origdb;
 	rc = nop->o_bd->be_search(nop, &nrs);
-	filter_free_x(nop, nop->ors_filter);
-	op->o_tmpfree( key, op->o_tmpmemctx );
+	filter_free_x(nop, nop->ors_filter, 1);
 
 	if(rc != LDAP_SUCCESS && rc != LDAP_NO_SUCH_OBJECT) {
 		op->o_bd->bd_info = (BackendInfo *) on->on_info;
 		send_ldap_error(op, rs, rc, "unique_search failed");
-		return(rs->sr_err);
-	}
+		rc = rs->sr_err;
+	} else if(uq.count) {
+		Debug(LDAP_DEBUG_TRACE, "=> unique_search found %d records\n", uq.count );
 
-	Debug(LDAP_DEBUG_TRACE, "=> unique_search found %d records\n", uq.count, 0, 0);
-
-	if(uq.count) {
+		errmsgsize = sizeof("non-unique attributes found with ") + key->bv_len;
+		errmsg = op->o_tmpalloc(errmsgsize, op->o_tmpmemctx);
+		snprintf( errmsg, errmsgsize, "non-unique attributes found with %s", key->bv_val );
 		op->o_bd->bd_info = (BackendInfo *) on->on_info;
-		send_ldap_error(op, rs, LDAP_CONSTRAINT_VIOLATION,
-			"some attributes not unique");
-		return(rs->sr_err);
+		send_ldap_error(op, rs, LDAP_CONSTRAINT_VIOLATION, errmsg);
+		op->o_tmpfree(errmsg, op->o_tmpmemctx);
+		rc = rs->sr_err;
+	} else {
+		Debug(LDAP_DEBUG_TRACE, "=> unique_search found no records\n" );
+		rc = SLAP_CB_CONTINUE;
 	}
 
-	return(SLAP_CB_CONTINUE);
+	op->o_tmpfree( key->bv_val, op->o_tmpmemctx );
+
+	return(rc);
+}
+
+static int
+unique_unlock(
+	Operation *op,
+	SlapReply *rs
+)
+{
+	slap_callback *sc = op->o_callback;
+	unique_data *private = sc->sc_private;
+
+	ldap_pvt_thread_mutex_unlock( &private->serial_mutex );
+	op->o_callback = sc->sc_next;
+	op->o_tmpfree( sc, op->o_tmpmemctx );
+	return 0;
 }
 
 static int
@@ -1013,25 +1059,49 @@ unique_add(
 	Operation nop = *op;
 	Attribute *a;
 	char *key, *kp;
+	struct berval bvkey;
 	int rc = SLAP_CB_CONTINUE;
+	int locked = 0;
 
 	Debug(LDAP_DEBUG_TRACE, "==> unique_add <%s>\n",
-	      op->o_req_dn.bv_val, 0, 0);
+	      op->o_req_dn.bv_val );
+
+	if ( SLAPD_SYNC_IS_SYNCCONN( op->o_connid ) || (
+			get_relax(op) > SLAP_CONTROL_IGNORED
+			&& access_allowed( op, op->ora_e,
+				slap_schema.si_ad_entry, NULL,
+				ACL_MANAGE, NULL ) ) ) {
+		return rc;
+	}
 
 	for ( domain = legacy ? legacy : domains;
 	      domain;
-	      domain = domain->next ) {
+	      domain = domain->next )
+	{
 		unique_domain_uri *uri;
-		int ks = 0;
 
 		for ( uri = domain->uri;
 		      uri;
-		      uri = uri->next ) {
+		      uri = uri->next )
+		{
+			int len;
+			int ks = 0;
 
-			if ( uri->ndn
-			     && !dnIsSuffix( &op->o_req_ndn, uri->ndn ))
+			if ( uri->ndn.bv_val
+			     && !dnIsSuffix( &op->o_req_ndn, &uri->ndn ))
 				continue;
 
+			if ( uri->f ) {
+				if ( test_filter( NULL, op->ora_e, uri->f )
+					== LDAP_COMPARE_FALSE )
+				{
+					Debug( LDAP_DEBUG_TRACE,
+						"==> unique_add_skip<%s>\n",
+						op->o_req_dn.bv_val );
+					continue;
+				}
+			}
+
 			if(!(a = op->ora_e->e_attrs)) {
 				op->o_bd->bd_info = (BackendInfo *) on->on_info;
 				send_ldap_error(op, rs, LDAP_INVALID_SYNTAX,
@@ -1051,13 +1121,26 @@ unique_add(
 			/* skip this domain-uri if it isn't involved */
 			if ( !ks ) continue;
 
-			if ( uri->filter && uri->filter->bv_len )
-				ks += uri->filter->bv_len + STRLENOF ("(&)");
+			if ( domain->serial && !locked ) {
+				ldap_pvt_thread_mutex_lock( &private->serial_mutex );
+				locked = 1;
+			}
+
+			/* terminating NUL */
+			ks += sizeof("(|)");
+
+			if ( uri->filter.bv_val && uri->filter.bv_len )
+				ks += uri->filter.bv_len + STRLENOF ("(&)");
 			kp = key = op->o_tmpalloc(ks, op->o_tmpmemctx);
 
-			if ( uri->filter && uri->filter->bv_len )
-				kp += sprintf (kp, "(&%s", uri->filter->bv_val);
-			kp += sprintf(kp, "(|");
+			if ( uri->filter.bv_val && uri->filter.bv_len ) {
+				len = snprintf (kp, ks, "(&%s", uri->filter.bv_val);
+				assert( len >= 0 && len < ks );
+				kp += len;
+			}
+			len = snprintf(kp, ks - (kp - key), "(|");
+			assert( len >= 0 && len < ks - (kp - key) );
+			kp += len;
 
 			for(a = op->ora_e->e_attrs; a; a = a->a_next)
 				kp = build_filter(domain,
@@ -1065,26 +1148,45 @@ unique_add(
 						  a->a_desc,
 						  a->a_vals,
 						  kp,
+						  ks - ( kp - key ),
 						  op->o_tmpmemctx);
 
-			kp += sprintf(kp, ")");
-			if ( uri->filter && uri->filter->bv_len )
-				kp += sprintf (kp, ")");
+			len = snprintf(kp, ks - (kp - key), ")");
+			assert( len >= 0 && len < ks - (kp - key) );
+			kp += len;
+			if ( uri->filter.bv_val && uri->filter.bv_len ) {
+				len = snprintf(kp, ks - (kp - key), ")");
+				assert( len >= 0 && len < ks - (kp - key) );
+				kp += len;
+			}
+			bvkey.bv_val = key;
+			bvkey.bv_len = kp - key;
 
 			rc = unique_search ( op,
 					     &nop,
-					     uri->ndn ?
-					     uri->ndn :
+					     uri->ndn.bv_val ?
+					     &uri->ndn :
 					     &op->o_bd->be_nsuffix[0],
 					     uri->scope,
 					     rs,
-					     key);
+					     &bvkey);
 
 			if ( rc != SLAP_CB_CONTINUE ) break;
 		}
 		if ( rc != SLAP_CB_CONTINUE ) break;
 	}
 
+	if ( locked ) {
+		if ( rc != SLAP_CB_CONTINUE ) {
+			ldap_pvt_thread_mutex_unlock( &private->serial_mutex );
+		} else {
+			slap_callback *cb = op->o_tmpcalloc( 1, sizeof(slap_callback), op->o_tmpmemctx );
+			cb->sc_cleanup = unique_unlock;
+			cb->sc_private = private;
+			cb->sc_next = op->o_callback;
+			op->o_callback = cb;
+		}
+	}
 	return rc;
 }
 
@@ -1102,53 +1204,83 @@ unique_modify(
 	unique_domain *domain;
 	Operation nop = *op;
 	Modifications *m;
+	Entry *e = NULL;
 	char *key, *kp;
+	struct berval bvkey;
 	int rc = SLAP_CB_CONTINUE;
+	int locked = 0;
 
 	Debug(LDAP_DEBUG_TRACE, "==> unique_modify <%s>\n",
-	      op->o_req_dn.bv_val, 0, 0);
+	      op->o_req_dn.bv_val );
+
+	if ( !op->orm_modlist ) {
+		Debug(LDAP_DEBUG_TRACE, "unique_modify: got empty modify op\n" );
+		return rc;
+	}
+
+	if ( SLAPD_SYNC_IS_SYNCCONN( op->o_connid ) || (
+			get_relax(op) > SLAP_CONTROL_IGNORED
+			&& overlay_entry_get_ov(op, &op->o_req_ndn, NULL, NULL, 0, &e, on) == LDAP_SUCCESS
+			&& e
+			&& access_allowed( op, e,
+				slap_schema.si_ad_entry, NULL,
+				ACL_MANAGE, NULL ) ) ) {
+		overlay_entry_release_ov( op, e, 0, on );
+		return rc;
+	}
+	if ( e ) {
+		overlay_entry_release_ov( op, e, 0, on );
+	}
 
 	for ( domain = legacy ? legacy : domains;
 	      domain;
-	      domain = domain->next ) {
+	      domain = domain->next )
+	{
 		unique_domain_uri *uri;
-		int ks = 0;
 
 		for ( uri = domain->uri;
 		      uri;
-		      uri = uri->next ) {
+		      uri = uri->next )
+		{
+			int len;
+			int ks = 0;
 
-			if ( uri->ndn
-			     && !dnIsSuffix( &op->o_req_ndn, uri->ndn ))
+			if ( uri->ndn.bv_val
+			     && !dnIsSuffix( &op->o_req_ndn, &uri->ndn ))
 				continue;
 
-			if ( !(m = op->orm_modlist) ) {
-				op->o_bd->bd_info = (BackendInfo *) on->on_info;
-				send_ldap_error(op, rs, LDAP_INVALID_SYNTAX,
-						"unique_modify() got null op.orm_modlist");
-				rc = rs->sr_err;
-				break;
-
-			} else
-				for ( ; m; m = m->sml_next)
-					if ( (m->sml_op & LDAP_MOD_OP)
-					     != LDAP_MOD_DELETE )
-						ks += count_filter_len
-							( domain,
-							  uri,
-							  m->sml_desc,
-							  m->sml_values);
+			for ( m = op->orm_modlist; m; m = m->sml_next)
+				if ( (m->sml_op & LDAP_MOD_OP)
+				     != LDAP_MOD_DELETE )
+					ks += count_filter_len
+						( domain,
+						  uri,
+						  m->sml_desc,
+						  m->sml_values);
 
 			/* skip this domain-uri if it isn't involved */
 			if ( !ks ) continue;
 
-			if ( uri->filter && uri->filter->bv_len )
-				ks += uri->filter->bv_len + STRLENOF ("(&)");
+			if ( domain->serial && !locked ) {
+				ldap_pvt_thread_mutex_lock( &private->serial_mutex );
+				locked = 1;
+			}
+
+			/* terminating NUL */
+			ks += sizeof("(|)");
+
+			if ( uri->filter.bv_val && uri->filter.bv_len )
+				ks += uri->filter.bv_len + STRLENOF ("(&)");
 			kp = key = op->o_tmpalloc(ks, op->o_tmpmemctx);
 
-			if ( uri->filter && uri->filter->bv_len )
-				kp += sprintf (kp, "(&%s", uri->filter->bv_val);
-			kp += sprintf(kp, "(|");
+			if ( uri->filter.bv_val && uri->filter.bv_len ) {
+				len = snprintf(kp, ks, "(&%s", uri->filter.bv_val);
+				assert( len >= 0 && len < ks );
+				kp += len;
+			}
+			len = snprintf(kp, ks - (kp - key), "(|");
+			assert( len >= 0 && len < ks - (kp - key) );
+			kp += len;
 
 			for(m = op->orm_modlist; m; m = m->sml_next)
 				if ( (m->sml_op & LDAP_MOD_OP)
@@ -1158,26 +1290,45 @@ unique_modify(
 							    m->sml_desc,
 							    m->sml_values,
 							    kp,
+							    ks - (kp - key),
 							    op->o_tmpmemctx );
 
-			kp += sprintf (kp, ")");
-			if ( uri->filter && uri->filter->bv_len )
-				kp += sprintf (kp, ")");
+			len = snprintf(kp, ks - (kp - key), ")");
+			assert( len >= 0 && len < ks - (kp - key) );
+			kp += len;
+			if ( uri->filter.bv_val && uri->filter.bv_len ) {
+				len = snprintf (kp, ks - (kp - key), ")");
+				assert( len >= 0 && len < ks - (kp - key) );
+				kp += len;
+			}
+			bvkey.bv_val = key;
+			bvkey.bv_len = kp - key;
 
 			rc = unique_search ( op,
 					     &nop,
-					     uri->ndn ?
-					     uri->ndn :
+					     uri->ndn.bv_val ?
+					     &uri->ndn :
 					     &op->o_bd->be_nsuffix[0],
 					     uri->scope,
 					     rs,
-					     key);
+					     &bvkey);
 
 			if ( rc != SLAP_CB_CONTINUE ) break;
 		}
 		if ( rc != SLAP_CB_CONTINUE ) break;
 	}
 
+	if ( locked ) {
+		if ( rc != SLAP_CB_CONTINUE ) {
+			ldap_pvt_thread_mutex_unlock( &private->serial_mutex );
+		} else {
+			slap_callback *cb = op->o_tmpcalloc( 1, sizeof(slap_callback), op->o_tmpmemctx );
+			cb->sc_cleanup = unique_unlock;
+			cb->sc_private = private;
+			cb->sc_next = op->o_callback;
+			op->o_callback = cb;
+		}
+	}
 	return rc;
 }
 
@@ -1194,29 +1345,48 @@ unique_modrdn(
 	unique_domain *legacy = private->legacy;
 	unique_domain *domain;
 	Operation nop = *op;
+	Entry *e = NULL;
 	char *key, *kp;
+	struct berval bvkey;
 	LDAPRDN	newrdn;
 	struct berval bv[2];
 	int rc = SLAP_CB_CONTINUE;
+	int locked = 0;
 
 	Debug(LDAP_DEBUG_TRACE, "==> unique_modrdn <%s> <%s>\n",
-		op->o_req_dn.bv_val, op->orr_newrdn.bv_val, 0);
+		op->o_req_dn.bv_val, op->orr_newrdn.bv_val );
+
+	if ( SLAPD_SYNC_IS_SYNCCONN( op->o_connid ) || (
+			get_relax(op) > SLAP_CONTROL_IGNORED
+			&& overlay_entry_get_ov(op, &op->o_req_ndn, NULL, NULL, 0, &e, on) == LDAP_SUCCESS
+			&& e
+			&& access_allowed( op, e,
+				slap_schema.si_ad_entry, NULL,
+				ACL_MANAGE, NULL ) ) ) {
+		overlay_entry_release_ov( op, e, 0, on );
+		return rc;
+	}
+	if ( e ) {
+		overlay_entry_release_ov( op, e, 0, on );
+	}
 
 	for ( domain = legacy ? legacy : domains;
 	      domain;
-	      domain = domain->next ) {
+	      domain = domain->next )
+	{
 		unique_domain_uri *uri;
-		int ks = 0;
 
 		for ( uri = domain->uri;
 		      uri;
-		      uri = uri->next ) {
-			int i;
+		      uri = uri->next )
+		{
+			int i, len;
+			int ks = 0;
 
-			if ( uri->ndn
-			     && !dnIsSuffix( &op->o_req_ndn, uri->ndn )
+			if ( uri->ndn.bv_val
+			     && !dnIsSuffix( &op->o_req_ndn, &uri->ndn )
 			     && (!op->orr_nnewSup
-				 || !dnIsSuffix( op->orr_nnewSup, uri->ndn )))
+				 || !dnIsSuffix( op->orr_nnewSup, &uri->ndn )))
 				continue;
 
 			if ( ldap_bv2rdn_x ( &op->oq_modrdn.rs_newrdn,
@@ -1259,13 +1429,26 @@ unique_modrdn(
 			/* skip this domain if it isn't involved */
 			if ( !ks ) continue;
 
-			if ( uri->filter && uri->filter->bv_len )
-				ks += uri->filter->bv_len + STRLENOF ("(&)");
+			if ( domain->serial && !locked ) {
+				ldap_pvt_thread_mutex_lock( &private->serial_mutex );
+				locked = 1;
+			}
+
+			/* terminating NUL */
+			ks += sizeof("(|)");
+
+			if ( uri->filter.bv_val && uri->filter.bv_len )
+				ks += uri->filter.bv_len + STRLENOF ("(&)");
 			kp = key = op->o_tmpalloc(ks, op->o_tmpmemctx);
 
-			if ( uri->filter && uri->filter->bv_len )
-				kp += sprintf (kp, "(&%s", uri->filter->bv_val);
-			kp += sprintf(kp, "(|");
+			if ( uri->filter.bv_val && uri->filter.bv_len ) {
+				len = snprintf(kp, ks, "(&%s", uri->filter.bv_val);
+				assert( len >= 0 && len < ks );
+				kp += len;
+			}
+			len = snprintf(kp, ks - (kp - key), "(|");
+			assert( len >= 0 && len < ks - (kp - key) );
+			kp += len;
 
 			for ( i=0; newrdn[i]; i++) {
 				bv[0] = newrdn[i]->la_value;
@@ -1274,27 +1457,46 @@ unique_modrdn(
 						    newrdn[i]->la_private,
 						    bv,
 						    kp,
+						    ks - (kp - key ),
 						    op->o_tmpmemctx);
 			}
 
-			kp += sprintf(kp, ")");
-			if ( uri->filter && uri->filter->bv_len )
-				kp += sprintf (kp, ")");
+			len = snprintf(kp, ks - (kp - key), ")");
+			assert( len >= 0 && len < ks - (kp - key) );
+			kp += len;
+			if ( uri->filter.bv_val && uri->filter.bv_len ) {
+				len = snprintf (kp, ks - (kp - key), ")");
+				assert( len >= 0 && len < ks - (kp - key) );
+				kp += len;
+			}
+			bvkey.bv_val = key;
+			bvkey.bv_len = kp - key;
 
 			rc = unique_search ( op,
 					     &nop,
-					     uri->ndn ?
-					     uri->ndn :
+					     uri->ndn.bv_val ?
+					     &uri->ndn :
 					     &op->o_bd->be_nsuffix[0],
 					     uri->scope,
 					     rs,
-					     key);
+					     &bvkey);
 
 			if ( rc != SLAP_CB_CONTINUE ) break;
 		}
 		if ( rc != SLAP_CB_CONTINUE ) break;
 	}
 
+	if ( locked ) {
+		if ( rc != SLAP_CB_CONTINUE ) {
+			ldap_pvt_thread_mutex_unlock( &private->serial_mutex );
+		} else {
+			slap_callback *cb = op->o_tmpcalloc( 1, sizeof(slap_callback), op->o_tmpmemctx );
+			cb->sc_cleanup = unique_unlock;
+			cb->sc_private = private;
+			cb->sc_next = op->o_callback;
+			op->o_callback = cb;
+		}
+	}
 	return rc;
 }
 
@@ -1312,10 +1514,9 @@ unique_initialize()
 	memset (&unique, 0, sizeof(unique));
 
 	unique.on_bi.bi_type = "unique";
+	unique.on_bi.bi_flags = SLAPO_BFLAG_SINGLE;
 	unique.on_bi.bi_db_init = unique_db_init;
 	unique.on_bi.bi_db_destroy = unique_db_destroy;
-	unique.on_bi.bi_db_open = unique_open;
-	unique.on_bi.bi_db_close = unique_close;
 	unique.on_bi.bi_op_add = unique_add;
 	unique.on_bi.bi_op_modify = unique_modify;
 	unique.on_bi.bi_op_modrdn = unique_modrdn;
